import numpy as np
import pandas as pd
import math
from ...functions.SOLWEIGpython.wall_cover import get_wall_cover
from .cylindric_wedge import cylindric_wedge_voxel
from .Lvikt_veg import Lvikt_veg

# Stefan Boltzmans Constant
SBC = 5.67051e-8

def load_walls(voxelTable, solweig_parameters, wall_type, wallaspect, Ta, timeStep, albedo_grid, landcover, lcgrid, dsm):
    '''This function loads the voxel data created in sky view factor calculator and turns it into a Pandas DataFrame.'''
    
    # Load data as Pandas DataFrame and add column names
    voxelTable = pd.DataFrame(voxelTable, columns = ['voxelId', 'voxelHeight', 'wallHeight', 'wallHeight_exact', 'wallId', 'ypos', 'xpos', 
                                                        'SVF_height', 'SVF', 'SVF_fix', 'svfbu', 'svfveg', 'svfaveg'])
    
    # Initiate/declare new columns used by SOLWEIG and parameterization scheme for wall surface temperatures
    voxelTable['wallTemperature'] = Ta # Initial wall temperature is set to air temperature
    voxelTable['timeStep'] = timeStep
    
    # Add columns filled later
    columns_to_add = ['SVF_ground', 'svfbu_at_ground', 'svfaveg_at_ground', 'wallAspect', 'wallEmissivity', 'wallThickness', 
                      'wallAlbedo', 'thermalEffusivity', 'thermalDiffusivity', 'groundAlbedo', 'wallShade', 'wallShadeHeight', 
                      'LongwaveRadiation', 'K_in', 'L_in', 'Lwallsun', 'Lwallsh', 'Lrefl', 'Lveg', 'Lground', 'Lsky', 'esky', 'voxelHeightMasl']
    for col in columns_to_add:
        voxelTable[col] = 0
    
    
    # tmp = svf + svfveg - 1.
    # tmp[tmp < 0.] = 0.
    # %matlab crazyness around 0
    # svfalfa = np.arcsin(np.exp((np.log((1. - tmp)) / 2.)))
    tmp = voxelTable['SVF_fix'].to_numpy() + voxelTable['svfveg'].to_numpy() - 1.
    tmp[tmp < 0.] = 0.
    voxelTable['svfalfa'] = np.arcsin(np.exp((np.log((1. - tmp)) / 2.)))

    # Add wall aspect of each voxel
    # temp_table = np.column_stack([voxelTable['voxelId'].to_numpy(), voxelTable['ypos'].to_numpy(), voxelTable['xpos'].to_numpy()])
    temp_table = np.column_stack([voxelTable['wallId'].to_numpy(), voxelTable['ypos'].to_numpy(), voxelTable['xpos'].to_numpy()])
    temp_table = np.unique(temp_table, axis = 0)
    # Add wall aspect
    for i in np.arange(temp_table.shape[0]):
        temp_aspect = wallaspect[temp_table[i,1].astype(int), temp_table[i,2].astype(int)]
        voxelTable.loc[voxelTable['wallId'] == temp_table[i,0], 'wallAspect'] = temp_aspect
        temp_building = voxelTable.loc[((voxelTable['wallId'] == temp_table[i,0]) & (voxelTable['voxelHeight'] == voxelTable['voxelHeight'].min())), 'svfbu'].copy().to_numpy()[0]
        temp_veg = voxelTable.loc[((voxelTable['wallId'] == temp_table[i,0]) & (voxelTable['voxelHeight'] == voxelTable['voxelHeight'].min())), 'svfaveg'].copy().to_numpy()[0]
        temp_albedo = albedo_grid[temp_table[i,1].astype(int), temp_table[i,2].astype(int)]
        temp_dsm = dsm[temp_table[i,1].astype(int), temp_table[i,2].astype(int)]
        
        voxelTable.loc[voxelTable['wallId'] == temp_table[i,0], 'svfbu_at_ground'] = temp_building
        voxelTable.loc[voxelTable['wallId'] == temp_table[i,0], 'svfaveg_at_ground'] = temp_veg
        voxelTable.loc[voxelTable['wallId'] == temp_table[i,0], 'groundAlbedo'] = temp_albedo
        voxelTable.loc[voxelTable['wallId'] == temp_table[i,0], 'voxelHeightMasl'] = voxelTable.loc[voxelTable['wallId'] == temp_table[i,0], 'voxelHeight'].to_numpy() + temp_dsm
    
    # Set voxelId to index
    voxelTable = voxelTable.set_index('voxelId')

    # Calculate fractions
    # voxelTable.loc[voxelTable['svfbu'] > 0.5, 'svfbu'] = 0.5
    # voxelTable.loc[voxelTable['svfaveg'] > 0.5, 'svfaveg'] = 0.5
    # Non-sky fraction = (1 - svf at ground level) - 0.5 (ground surface seen from wall)
    building_fraction = 1 - voxelTable['svfbu_at_ground'].to_numpy() - 0.5
    building_fraction[building_fraction < 0] = 0.
    veg_fraction = 1 - voxelTable['svfaveg_at_ground'].to_numpy() - 0.5
    veg_fraction[veg_fraction < 0] = 0.
    voxelTable['building_fraction'] = building_fraction
    voxelTable['veg_fraction'] = veg_fraction
    sky_fraction = voxelTable['SVF_fix'].to_numpy()
    ground_fraction = 1 - sky_fraction - building_fraction - veg_fraction
    voxelTable['ground_fraction'] = ground_fraction

    voxelTable['total_fraction'] = building_fraction + sky_fraction + ground_fraction + veg_fraction

    # Set thermal effusivity according to wall type (either from GUI or from landcover)
    if landcover == 1:
        unique_landcover = np.unique(lcgrid)
        # Unique wall codes for unique wall types, e.g. brick, concrete, wood, etc.
        unique_walls = unique_landcover[unique_landcover > 99].astype(int)
        # Get wall properties for wall surface temperature scheme
        if unique_walls.size > 1:
            voxelTable = get_wall_cover(voxelTable, lcgrid, dsm, solweig_parameters)
        elif unique_walls.size == 1:
            # Specific heat capacity
            wallTc = solweig_parameters['Specific_heat']['Value'][solweig_parameters['Names']['Value'][str(unique_walls[0])]]
            # Thermal conductivity
            wallTk = solweig_parameters['Thermal_conductivity']['Value'][solweig_parameters['Names']['Value'][str(unique_walls[0])]]
            # Material density
            wallD = solweig_parameters['Density']['Value'][solweig_parameters['Names']['Value'][str(unique_walls[0])]]
            # Thermal effusivity
            wallTu = np.sqrt(wallTc * wallD * wallTk)
            # Thermal diffusivity
            wallTd = wallTk/(wallTc*wallD)
            # Set thermal effusivity
            voxelTable['thermalEffusivity'] = wallTu      
            # Set thermal diffusivity
            voxelTable['thermalDiffusivity'] = wallTd
            # Set wall albedo
            voxelTable['wallAlbedo'] = solweig_parameters['Albedo']['Material']['Value'][solweig_parameters['Names']['Value'][str(unique_walls[0])]]      
            # Set wall emissivity
            voxelTable['wallEmissivity'] = solweig_parameters['Emissivity']['Value'][solweig_parameters['Names']['Value'][str(unique_walls[0])]]                  
            # Set thickness
            voxelTable['wallThickness'] = solweig_parameters['Wall_thickness']['Value'][solweig_parameters['Names']['Value'][str(unique_walls[0])]]
        else:
            landcover = 0
    
    if landcover == 0:
        # Specific heat capacity
        wallTc = solweig_parameters['Specific_heat']['Value'][solweig_parameters['Names']['Value'][wall_type]]
        # Thermal conductivity
        wallTk = solweig_parameters['Thermal_conductivity']['Value'][solweig_parameters['Names']['Value'][wall_type]]
        # Material density
        wallD = solweig_parameters['Density']['Value'][solweig_parameters['Names']['Value'][wall_type]]
        # Thermal effusivity
        wallTu = np.sqrt(wallTc * wallD * wallTk)    
        # Set thermal effusivity
        voxelTable['thermalEffusivity'] = wallTu
        # Calculate thermal diffusivity
        wallTd = wallTk/(wallTc*wallD)
        # Set thermal diffusivity
        voxelTable['thermalDiffusivity'] = wallTd
        # Get wall albedo
        voxelTable['wallAlbedo'] = solweig_parameters['Albedo']['Material']['Value'][solweig_parameters['Names']['Value'][wall_type]]
        # Get wall emissivity
        voxelTable['wallEmissivity'] = solweig_parameters['Emissivity']['Value'][solweig_parameters['Names']['Value'][wall_type]]
        # Get wall thickness
        voxelTable['wallThickness'] = solweig_parameters['Wall_thickness']['Value'][solweig_parameters['Names']['Value'][wall_type]]

    gvc_sim = False
    kolumbus_sim = True
    eqTime = True
    if gvc_sim:
        # Changed aspect on GVC walls
        print('wall aspect for wall 919 before change:')
        print(voxelTable.loc[voxelTable['wallId'] == 919, 'wallAspect'])
        print('wall aspect for wall 925 before change:')
        print(voxelTable.loc[voxelTable['wallId'] == 925, 'wallAspect'])

        east_angle = 130
        west_angle = 210
        # print('wall aspect for wall 919 after change:')
        voxelTable.loc[voxelTable['wallId'] == 914, 'wallAspect'] = east_angle # 914 = east
        voxelTable.loc[voxelTable['wallId'] == 920, 'wallAspect'] = east_angle # 920 = southeast
        voxelTable.loc[voxelTable['wallId'] == 926, 'wallAspect'] = east_angle # 926 = south/southeast
        voxelTable.loc[voxelTable['wallId'] == 919, 'wallAspect'] = west_angle # 919 = west        
        voxelTable.loc[voxelTable['wallId'] == 925, 'wallAspect'] = west_angle # 925 = southwest
        
        print('wall aspect for wall 920 after change:')
        print(voxelTable.loc[voxelTable['wallId'] == 920, 'wallAspect'])
        print('wall aspect for wall 925 after change:')
        print(voxelTable.loc[voxelTable['wallId'] == 925, 'wallAspect'])
        wall_id = np.array([914, 919, 920, 925, 926])
        aspect_dict = {914: 135, 919: 220, 920: 135, 925: 220, 926: 135}
        for temp_id in wall_id:
            temp_y = voxelTable.loc[voxelTable['wallId'] == temp_id, 'ypos'].to_numpy().astype(int)
            temp_x = voxelTable.loc[voxelTable['wallId'] == temp_id, 'xpos'].to_numpy().astype(int)
            wallaspect[temp_y, temp_x] = aspect_dict[temp_id]

    elif kolumbus_sim == True:
        voxelTable.loc[voxelTable['voxelHeight'] >= 4, 'thermalEffusivity'] = 472
        voxelTable.loc[voxelTable['voxelHeight'] >= 4, 'wallAlbedo'] = 0.6
        voxelTable.loc[voxelTable['voxelHeight'] >= 4, 'wallEmissivity'] = 0.95
        voxelTable.loc[voxelTable['voxelHeight'] >= 4, 'timeStep'] = 0.03**2/(np.pi**2*0.13*10**-6)

        voxelTable.loc[voxelTable['voxelHeight'] < 4, 'thermalEffusivity'] = (1068+860)/2
        voxelTable.loc[voxelTable['voxelHeight'] < 4, 'wallAlbedo'] = 0.5
        voxelTable.loc[voxelTable['voxelHeight'] < 4, 'wallEmissivity'] = 0.95
        voxelTable.loc[voxelTable['voxelHeight'] < 4, 'timeStep'] = 0.1**2/(np.pi**2*((0.62+0.44)/2)*10**-6)
        # voxelTable['thermalEffusivity'] = 472
        # voxelTable['wallAlbedo'] = 0.52
        # voxelTable['wallEmissivity'] = 0.95
        # voxelTable['timeStep'] = 0.03**2/(np.pi**2*0.13*10**-6)

        # voxelTable['thermalEffusivity'] = (1068+860)/2
        # voxelTable['wallAlbedo'] = 0.52
        # voxelTable['wallEmissivity'] = 0.95
        # voxelTable['timeStep'] = 0.1**2/(np.pi**2*((0.62+0.44)/2)*10**-6)        

    if eqTime:
        voxelTable['timeStep'] = voxelTable['wallThickness'].to_numpy()**2/(np.pi**2 * voxelTable['thermalDiffusivity'].to_numpy())

    return voxelTable, wallaspect

def step_heating(q, e, t):
    '''Function to calculate delta surface temperature based on heat flux (q), thermal effusivity (e) and time (t)'''
    return (2*q)/e*np.sqrt(t/np.pi)

def surface_temperature_calc(effusivity, t, Kin, Lin, Ta, wall_emissivity, Ts_previous):
    '''Function to get surface temperature'''
       
    dT = np.zeros((effusivity.shape[0]))
    Ts = np.zeros((effusivity.shape[0]))
    # Calculate heat flux based on Ts in previous time step and Kin, Lin and Ta from current time step
    Lout_temp = wall_emissivity * SBC * (Ts_previous + 273.15)**4
    # sensible_temp = 20 * (Ts_previous - Ta)
    energy_in_temp = Kin + Lin - Lout_temp #- sensible_temp
    # Calculate dT (Ts - Ta) with step wise heating
    dT = step_heating(energy_in_temp, effusivity, t)

    # Surface temperature
    Ts_current = Ta + dT
    # Estimate surface temperature with Ts of current timestep, i.e. not Ts_previous
    Lout_temp = wall_emissivity * SBC * (Ts_current + 273.15)**4
    energy_in_temp = Kin + Lin - Lout_temp #- sensible_temp
    dT = step_heating(energy_in_temp, effusivity, t)
    # Surface temperature after second iteration
    Ts = Ta + dT
        
    return Ts, dT

def wall_surface_temperature(voxelTable, wallsh, altitude, azimuth, timeStep, K_direct, K_diff, K_down, Ldown, Lup, Ta, esky):
    '''Wall surface temperature parameterization
    
    This parameterization scheme estimates a wall temperature based on the 
    incoming energy (short- and longwave radiation), outgoing longwave radiation and sensible heat.'''

    # Stefan Boltzmans Constant
    SBC = 5.67051e-8

    # Degrees fo radians conversion factor
    deg2rad = np.pi/180

    # Estimate shadow on wall, i.e. how far up the wall that the shade stretches (or how far down the wall that the sun reaches)
    voxelTable['wallShade'] = 0 # Starting value is zero, i.e. the entire wall is shaded
    # If sun is above horizon, get wall shadow height on wall from wallsh calculated in shadowingfunction_wallheight_23/13 in Solweig_2022a_calc_forprocessing.py
    if altitude > 0:
        # Starting value of wallShadeHeight
        voxelTable['wallShadeHeight'] = 0
        # Find number of unique walls in model domain
        unique_walls = np.unique(voxelTable['wallId'])
        # Find wall shade height of each unique wall
        for unique_wall in unique_walls:
            # temp_sh is the shadow height of a wall
            temp_sh = wallsh[voxelTable.loc[voxelTable.wallId == unique_wall, 'ypos'].to_numpy().astype(int)[0], voxelTable.loc[voxelTable.wallId == unique_wall, 'xpos'].to_numpy().astype(int)[0]]
            #print(temp_sh)
            # Change to sunlit (1) for voxels that are above wall shade height
            voxelTable.loc[(voxelTable['wallId'] == unique_wall) & (voxelTable['voxelHeight'] >= temp_sh) & (voxelTable['wallHeight_exact'] > temp_sh), 'wallShade'] = 1
            # Add wall shade height to pandas dataframe
            voxelTable.loc[(voxelTable['wallId'] == unique_wall), 'wallShadeHeight'] = temp_sh
    # If sun is below horizon, everything is in "shade"
    else:
        voxelTable['wallShadeHeight'] = voxelTable['wallHeight_exact']

    # Ldown and Lup for wall pixels used when estimating the amount of longwave received from surrounding surfaces (ground and reflected)
    Ldown_array = np.zeros((voxelTable.shape[0]))
    Lup_array = np.zeros((voxelTable.shape[0]))
    for idx in np.arange(voxelTable.shape[0]):
        temp_y = voxelTable.iloc[idx]['ypos'].astype(int)
        temp_x = voxelTable.iloc[idx]['xpos'].astype(int)
        Ldown_array[idx] = Ldown[temp_y, temp_x]
        Lup_array[idx] = Lup[temp_y, temp_x]

    # Estimate longwave in based on old methods by Bogren et al.
    # If sun above horizon
    if altitude > 0:
        # Cylindric wedge based on building height angle from svf and solar zenith angle
        F_sh = cylindric_wedge_voxel((90 - altitude) * deg2rad, voxelTable['svfalfa'].to_numpy())  # Fraction shadow on building walls based on sun alt and svf
        F_sh[np.isnan(F_sh)] = 0.5

        # Only half of hemisphere visible from a wall, therefore the entire seen area can be sunlit, compared to an open space where only half can be sunlit
        F_sh = 2. * F_sh - 1.  #(cylindric_wedge scaled 0-1)

        # Fraction of sunlit and shaded building surfaces seen by a building wall based on its aspect and the azimuth of the sun
        wallSun = np.abs(voxelTable['wallAspect'].to_numpy() - azimuth)
        wallSun = wallSun/180.
        wallSun[wallSun > 1.] = 2 - (wallSun[wallSun > 1.])
        wallSun = 0.2 + wallSun * 0.6 # Scaling wall shadows to avoid totally shaded and totally sunlit walls

        # Temperature in shade and sun based on mean values for all voxels, i.e. shade = mean of all shaded and sun = mean of all sunlit
        ts_shade = voxelTable.loc[voxelTable['wallShade'] == 0, 'wallTemperature'].to_numpy().mean()
        ts_sun = voxelTable.loc[voxelTable['wallShade'] == 1, 'wallTemperature'].to_numpy().mean()

        # Calculation of longwave radiation received by the voxels based on sunlit and shaded surroundings
        Lwallsun = (SBC * voxelTable['wallEmissivity'].to_numpy() * ((ts_sun + 273.15)**4) * voxelTable['building_fraction'].to_numpy() * (1. - F_sh)) * wallSun # (1 - svf - 0.5) istället för viktwall?
        Lwallsh = (SBC * voxelTable['wallEmissivity'].to_numpy() * ((ts_shade + 273.15)**4) * voxelTable['building_fraction'].to_numpy() * (1. - F_sh)) * (1 - wallSun) # (1 - svf - 0.5) istället för viktwall?
        
        Lwallsh += SBC * voxelTable['wallEmissivity'].to_numpy() * ((ts_shade + 273.15)**4) * voxelTable['building_fraction'].to_numpy() * F_sh # * 0.5 # (1 - svf - 0.5) istället för viktwall?

    # If sun below horizon
    else:
        Lwallsun = 0.
        ts_shade = voxelTable.loc[voxelTable['wallShade'] == 0, 'wallTemperature'].to_numpy().mean()
        Lwallsh = SBC * voxelTable['wallEmissivity'].to_numpy() * (ts_shade + 273.15)**4 * voxelTable['building_fraction'].to_numpy() # * 0.5 # (1 - svf - 0.5) istället för viktwall?

    # Received longwave radiation from vegetation
    Lveg = SBC * voxelTable['wallEmissivity'].to_numpy() * (Ta + 273.15)**4 * voxelTable['veg_fraction'].to_numpy() # * 0.5 # (1 - svf - 0.5) istället för viktwall?
    # Received longwave radiation from the sky
    Lsky = SBC * esky * ((Ta + 273.15)**4) * voxelTable['SVF_fix'].to_numpy() # * 0.5 # svf istället för viktwall?
    # Received reflected longwave radiation
    Lrefl = (1. - voxelTable['wallEmissivity'].to_numpy()) * (Ldown_array + Lup_array) * voxelTable['building_fraction'].to_numpy() # * 0.5 # (1 - svf - 0.5) istället för viktwall?
    # Received longwave radiation from ground
    Lground = Lup_array * voxelTable['ground_fraction'].to_numpy()
    
    # Total amount of longwave radiation received by a wall surface
    L_in = Lwallsun + Lwallsh + Lrefl + Lveg + Lground + Lsky #TODO ground??
    
    voxelTable['Lwallsun'] = Lwallsun
    voxelTable['Lwallsh'] = Lwallsh
    voxelTable['Lrefl'] = Lrefl
    voxelTable['Lveg'] = Lveg
    voxelTable['Lground'] = Lground
    voxelTable['Lsky'] = Lsky
    voxelTable['esky'] = esky
    if altitude > 0:
        voxelTable['F_sh'] = F_sh
        voxelTable['wallSun'] = wallSun
    else:
        voxelTable['F_sh'] = 0.
        voxelTable['wallSun'] = 0.

    # Angle of incidence (wall aspect vs solar position (altitude and azimuth))
    sun_x = 1 * np.cos(math.radians(altitude)) * np.cos(math.radians(azimuth)) * np.cos(deg2rad * voxelTable['wallAspect'].to_numpy())
    sun_y = 1 * np.cos(math.radians(altitude)) * np.sin(math.radians(azimuth)) * np.sin(deg2rad * voxelTable['wallAspect'].to_numpy())
    cf = sun_x + sun_y
    cf[cf <= 0] = 0.

    # If sun above horizon
    if altitude > 0:
        K_in = (1 - voxelTable['wallAlbedo'].to_numpy()) * (K_direct * cf * voxelTable['wallShade'].to_numpy() + 
                                                      K_diff * voxelTable['SVF_fix'].to_numpy() + 
                                                       K_down * voxelTable['wallAlbedo'].to_numpy() * voxelTable['building_fraction'] + 
                                                       (K_down * voxelTable['groundAlbedo']) * voxelTable['ground_fraction'])
    # If sun below horizon
    else:
        K_in = 0.    

    voxelTable['K_in'] = K_in
    voxelTable['L_in'] = L_in

    # Calculate wall surface temperature with step-wise heating
    Ts_previous = voxelTable['wallTemperature'].to_numpy()
    if voxelTable['timeStep'].unique().size == 1:
        timeStep = voxelTable.iloc[0]['timeStep']
    else:
        timeStep = voxelTable['timeStep'].to_numpy()
    Ts, dT = surface_temperature_calc(voxelTable['thermalEffusivity'].to_numpy(), timeStep, K_in, L_in, Ta, voxelTable['wallEmissivity'].to_numpy(), Ts_previous)
    voxelTable['wallTemperature'] = Ts
        
    # Convert wall temperature to longwave radiation
    voxelTable['LongwaveRadiation'] = ((voxelTable['wallEmissivity'] * SBC * ((voxelTable['wallTemperature'] + 273.15) ** 4)) / np.pi)

    # Position of the sun for current timestep (can be removed?)
    voxelTable['sunAltitude'] = altitude
    voxelTable['sunAzimuth'] = azimuth

    return voxelTable


import numpy as np

def get_wall_cover(voxelTable, lcgrid, dsm, lc_params):
    '''Function to set thermal properties of wall used in surface temperature scheme of walls'''

    # Y-position of wall pixel in raster
    ypos = voxelTable['ypos'].to_numpy().astype(int)
    # X-position of wall pixel in raster
    xpos = voxelTable['xpos'].to_numpy().astype(int)
    # Empty array to store wall code
    wallCode = np.zeros((voxelTable.shape[0]))
    # Empty array to store thermal effusivity
    wallTu = np.zeros((voxelTable.shape[0]))
    # Empty array to store thermal conductivity
    wallTd = np.zeros((voxelTable.shape[0]))
    # Empty array to store albedo
    wallAlbedo = np.zeros((voxelTable.shape[0]))
    # Empty array to store emissivity
    wallEmissivity = np.zeros((voxelTable.shape[0]))
    # Empty array to store wall thickness
    wallThickness = np.zeros((voxelTable.shape[0]))    
    # Search kernel used to find wall code
    domain = np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]])
    # Loop through all wall pixels in voxel table
    for i in range(voxelTable.shape[0]):
        # Temporary lc_grid based on kernel and wall y and x position
        temp_lc = lcgrid[ypos[i]-1:ypos[i]+2, xpos[i]-1:xpos[i]+2] * domain
        # Temporary dsm based on kernel and wall y and x position
        temp_dsm = dsm[ypos[i]-1:ypos[i]+2, xpos[i]-1:xpos[i]+2] * domain
        # Temporary code based on highest pixel in temp_dsm where temp_lc is a building
        temp_code = temp_lc[((temp_lc > 99) & (temp_dsm == temp_dsm.max()))]

        # If more than one option in temp_code, use the first one #TODO CHANGE TO MOST COMMON
        if len(temp_code) > 1:
            temp_code = temp_code[0].astype(int)
        elif len(temp_code) == 1:
            temp_code = temp_code[0].astype(int)
        # If no wall type specified in land cover for specific wall pixel, set to concrete
        elif temp_code.size == 0:
            temp_code = 101

        # Save wall code in array
        wallCode[i] = temp_code
        # Volumetric heat capacity
        temp_Tc = lc_params['Specific_heat']['Value'][lc_params['Names']['Value'][str(temp_code)]]
        # Thermal conductivity
        temp_Tk = lc_params['Thermal_conductivity']['Value'][lc_params['Names']['Value'][str(temp_code)]]
        # Material density
        temp_D = lc_params['Density']['Value'][lc_params['Names']['Value'][str(temp_code)]]
        # Calculate thermal effusivity
        temp_Tu = np.sqrt(temp_Tc * temp_D * temp_Tk)
        # Calculate thermal diffusivity
        temp_Td = temp_Tk/(temp_Tc*temp_D)
        # Save thermal effusivity of wall pixel in array
        wallTu[i] = temp_Tu
        # Save thermal diffusivity of wall pixel in array
        wallTd[i] = temp_Td
        # Get wall albedo
        temp_a = lc_params['Albedo']['Material']['Value'][lc_params['Names']['Value'][str(temp_code)]]
        # Save wall albedo
        wallAlbedo[i] = temp_a
        # Get wall emissivity
        temp_e = lc_params['Emissivity']['Value'][lc_params['Names']['Value'][str(temp_code)]]
        # Save wall emissivity
        wallEmissivity[i] = temp_e
        # Get wall thickness
        temp_wt = lc_params['Wall_thickness']['Value'][lc_params['Names']['Value'][str(temp_code)]]
        # Save wall thickness
        wallThickness[i] = temp_wt

    voxelTable['thermalEffusivity'] = wallTu
    voxelTable['thermalDiffusivity'] = wallTd
    voxelTable['wallAlbedo'] = wallAlbedo
    voxelTable['wallEmissivity'] = wallEmissivity
    voxelTable['wallThickness'] = wallThickness

    return voxelTable



__author__ = 'xlinfr'
import numpy as np


def sun_distance(jday):
    """

    #% Calculatesrelative earth sun distance
    #% with day of year as input.
    #% Partridge and Platt, 1975
    """
    b = 2.*np.pi*jday/365.
    D = np.sqrt((1.00011+np.dot(0.034221, np.cos(b))+np.dot(0.001280, np.sin(b))+np.dot(0.000719,
                                        np.cos((2.*b)))+np.dot(0.000077, np.sin((2.*b)))))
    return D



# %%
from pathlib import Path

import geopandas as gpd
from pyproj import CRS

from umep import (
    common,
    skyviewfactor_algorithm,
    wall_heightaspect_algorithm,
)
from umep.functions.SOLWEIGpython import solweig_runner_core

# working folder
input_folder = "demos/data/athens"
input_path = Path(input_folder).absolute()
input_path.mkdir(parents=True, exist_ok=True)
input_path_str = str(input_path)
# output folder
output_folder = "temp/athens"
output_folder_path = Path(output_folder).absolute()
output_folder_path.mkdir(parents=True, exist_ok=True)
output_folder_path_str = str(output_folder_path)
# extents
total_extents = [476800, 4205850, 477200, 4206250]

# %%
# buffer
working_crs = 2100
trees_gdf = gpd.read_file(input_folder + "/trees.gpkg")
trees_gdf = trees_gdf.to_crs(working_crs)
cdsm_rast, cdsm_transf = common.rasterise_gdf(
    trees_gdf,
    "geometry",
    "height",
    bbox=total_extents,
    pixel_size=1.0,
)
# add to DEM then set
common.save_raster(
    str(output_folder_path / "CDSM.tif"),
    cdsm_rast,
    cdsm_transf.to_gdal(),
    CRS.from_epsg(working_crs).to_wkt(),
)
# %%
# wall info for SOLWEIG
wall_heightaspect_algorithm.generate_wall_hts(
    dsm_path=input_path_str + "/DSM.tif",
    bbox=total_extents,
    out_dir=output_folder_path_str + "/walls",
)

# %%
# skyview factor for SOLWEIG
skyviewfactor_algorithm.generate_svf(
    dsm_path=input_path_str + "/DSM.tif",
    bbox=total_extents,
    out_dir=output_folder_path_str + "/svf",
    dem_path=input_path_str + "/DEM.tif",
    cdsm_path=output_folder_path_str + "/CDSM.tif",
    trans_veg_perc=3,
)

# %%
SRC = solweig_runner_core.SolweigRunCore(
    "demos/data/athens/configsolweig.ini",
    "demos/data/athens/parametersforsolweig.json",
)
SRC.run()



import numpy as np

def Kup_veg_2015a(radI,radD,radG,altitude,svfbuveg,albedo_b,F_sh,gvfalb,gvfalbE,gvfalbS,gvfalbW,gvfalbN,gvfalbnosh,gvfalbnoshE,gvfalbnoshS,gvfalbnoshW,gvfalbnoshN):

    Kup=(gvfalb*radI*np.sin(altitude*(np.pi/180.)))+(radD*svfbuveg+albedo_b*(1-svfbuveg)*(radG*(1-F_sh)+radD*F_sh))*gvfalbnosh

    KupE=(gvfalbE*radI*np.sin(altitude*(np.pi/180.)))+(radD*svfbuveg+albedo_b*(1-svfbuveg)*(radG*(1-F_sh)+radD*F_sh))*gvfalbnoshE

    KupS=(gvfalbS*radI*np.sin(altitude*(np.pi/180.)))+(radD*svfbuveg+albedo_b*(1-svfbuveg)*(radG*(1-F_sh)+radD*F_sh))*gvfalbnoshS

    KupW=(gvfalbW*radI*np.sin(altitude*(np.pi/180.)))+(radD*svfbuveg+albedo_b*(1-svfbuveg)*(radG*(1-F_sh)+radD*F_sh))*gvfalbnoshW

    KupN=(gvfalbN*radI*np.sin(altitude*(np.pi/180.)))+(radD*svfbuveg+albedo_b*(1-svfbuveg)*(radG*(1-F_sh)+radD*F_sh))*gvfalbnoshN

    return Kup, KupE, KupS, KupW, KupN


import numpy as np

from umep.util.SEBESOLWEIGCommonFiles.Perez_v3 import Perez_v3


def test_perez():
    radD = 200
    radI = 800
    jday = 100

    for zen_deg in range(0, 91, 5):
        for azi_deg in range(0, 361, 30):
            lv, _, _ = Perez_v3(zen_deg, azi_deg, radD, radI, jday, 1, 2)
            print("LV >>>", lv.min(), lv.mean(), lv.max())
            assert (~np.isnan(lv)).all()



from __future__ import absolute_import
author = 'xlinfr'

from . import sun_distance
import numpy as np
import math

def clearnessindex_2013b(zen, jday, Ta, RH, radG, location, P):

    """ Clearness Index at the Earth's surface calculated from Crawford and Duchon 1999

    :param zen: zenith angle in radians
    :param jday: day of year
    :param Ta: air temperature
    :param RH: relative humidity
    :param radG: global shortwave radiation
    :param location: distionary including lat, lon and alt
    :param P: pressure
    :return:
    """

    if P == -999.0:
        p = 1013.  # Pressure in millibars
    else:
        p = P*10.  # Convert from hPa to millibars

    Itoa = 1370.0  # Effective solar constant
    D = sun_distance.sun_distance(jday)  # irradiance differences due to Sun-Earth distances
    m = 35. * np.cos(zen) * ((1224. * (np.cos(zen)**2) + 1) ** (-1/2.))     # optical air mass at p=1013
    Trpg = 1.021-0.084*(m*(0.000949*p+0.051))**0.5  # Transmission coefficient for Rayliegh scattering and permanent gases

    # empirical constant depending on latitude
    if location['latitude'] < 10.:
        G = [3.37, 2.85, 2.80, 2.64]
    elif location['latitude'] >= 10. and location['latitude'] < 20.:
        G = [2.99, 3.02, 2.70, 2.93]
    elif location['latitude'] >= 20. and location['latitude'] <30.:
        G = [3.60, 3.00, 2.98, 2.93]
    elif location['latitude'] >= 30. and location['latitude'] <40.:
        G = [3.04, 3.11, 2.92, 2.94]
    elif location['latitude'] >= 40. and location['latitude'] <50.:
        G = [2.70, 2.95, 2.77, 2.71]
    elif location['latitude'] >= 50. and location['latitude'] <60.:
        G = [2.52, 3.07, 2.67, 2.93]
    elif location['latitude'] >= 60. and location['latitude'] <70.:
        G = [1.76, 2.69, 2.61, 2.61]
    elif location['latitude'] >= 70. and location['latitude'] <80.:
        G = [1.60, 1.67, 2.24, 2.63]
    elif location['latitude'] >= 80. and location['latitude'] <90.:
        G = [1.11, 1.44, 1.94, 2.02]

    if jday > 335 or jday <= 60:
        G = G[0]
    elif jday > 60 and jday <= 152:
        G = G[1]
    elif jday > 152 and jday <= 244:
        G = G[2]
    elif jday > 244 and jday <= 335:
        G = G[3]

    # dewpoint calculation
    a2 = 17.27
    b2 = 237.7
    Td = (b2*(((a2*Ta)/(b2+Ta))+np.log(RH)))/(a2-(((a2*Ta)/(b2+Ta))+np.log(RH)))
    Td = (Td*1.8)+32  # Dewpoint (F)
    u = np.exp(0.1133-np.log(G+1)+0.0393*Td)  # Precipitable water
    Tw = 1-0.077*((u*m)**0.3)  # Transmission coefficient for water vapor
    Tar = 0.935**m  # Transmission coefficient for aerosols

    I0=Itoa*np.cos(zen)*Trpg*Tw*D*Tar
    if abs(zen)>np.pi/2:
        I0 = 0
    # b=I0==abs(zen)>np.pi/2
    # I0(b==1)=0
    # clear b;
    if not(np.isreal(I0)):
        I0 = 0

    corr=0.1473*np.log(90-(zen/np.pi*180))+0.3454  # 20070329

    CIuncorr = radG / I0
    CI = CIuncorr + (1-corr)
    I0et = Itoa*np.cos(zen)*D  # extra terrestial solar radiation
    Kt = radG / I0et
    if math.isnan(CI):
        CI = float('Inf')

    return I0, CI, Kt, I0et, CIuncorr


# Ready for python action!
import numpy as np

# import matplotlib.pylab as plt
# from numba import njit, types


def shadowingfunctionglobalradiation(a, azimuth, altitude, scale, forsvf):
    # %This m.file calculates shadows on a DEM
    # % conversion
    degrees = np.pi / 180.0
    # if azimuth == 0.0:
    # azimuth = 0.000000000001
    azimuth = np.dot(azimuth, degrees)
    altitude = np.dot(altitude, degrees)
    # % measure the size of the image
    sizex = a.shape[0]
    sizey = a.shape[1]
    if forsvf == 0:
        barstep = np.max([sizex, sizey])
        total = 100.0 / barstep  # dlg.progressBar.setRange(0, barstep)
    # % initialise parameters
    f = a
    dx = 0.0
    dy = 0.0
    dz = 0.0
    temp = np.zeros((sizex, sizey))
    index = 1.0
    # % other loop parameters
    amaxvalue = a.max()
    pibyfour = np.pi / 4.0
    threetimespibyfour = 3.0 * pibyfour
    fivetimespibyfour = 5.0 * pibyfour
    seventimespibyfour = 7.0 * pibyfour
    sinazimuth = np.sin(azimuth)
    cosazimuth = np.cos(azimuth)
    tanazimuth = np.tan(azimuth)
    signsinazimuth = np.sign(sinazimuth)
    signcosazimuth = np.sign(cosazimuth)
    dssin = np.abs(1.0 / sinazimuth)
    dscos = np.abs(1.0 / cosazimuth)
    tanaltitudebyscale = np.tan(altitude) / scale
    # % main loop
    while amaxvalue >= dz and np.abs(dx) < sizex and np.abs(dy) < sizey:
        # while np.logical_and(np.logical_and(amaxvalue >= dz, np.abs(dx) <= sizex), np.abs(dy) <= sizey):(np.logical_and(amaxvalue >= dz, np.abs(dx) <= sizex), np.abs(dy) <= sizey):
        # if np.logical_or(np.logical_and(pibyfour <= azimuth, azimuth < threetimespibyfour), np.logical_and(fivetimespibyfour <= azimuth, azimuth < seventimespibyfour)):
        if (
            pibyfour <= azimuth
            and azimuth < threetimespibyfour
            or fivetimespibyfour <= azimuth
            and azimuth < seventimespibyfour
        ):
            dy = signsinazimuth * index
            dx = -1.0 * signcosazimuth * np.abs(np.round(index / tanazimuth))
            ds = dssin
        else:
            dy = signsinazimuth * np.abs(np.round(index * tanazimuth))
            dx = -1.0 * signcosazimuth * index
            ds = dscos

        # % note: dx and dy represent absolute values while ds is an incremental value
        dz = ds * index * tanaltitudebyscale
        temp[0:sizex, 0:sizey] = 0.0
        absdx = np.abs(dx)
        absdy = np.abs(dy)
        xc1 = (dx + absdx) / 2.0 + 1.0
        xc2 = sizex + (dx - absdx) / 2.0
        yc1 = (dy + absdy) / 2.0 + 1.0
        yc2 = sizey + (dy - absdy) / 2.0
        xp1 = -((dx - absdx) / 2.0) + 1.0
        xp2 = sizex - (dx + absdx) / 2.0
        yp1 = -((dy - absdy) / 2.0) + 1.0
        yp2 = sizey - (dy + absdy) / 2.0
        temp[int(xp1) - 1 : int(xp2), int(yp1) - 1 : int(yp2)] = (
            a[int(xc1) - 1 : int(xc2), int(yc1) - 1 : int(yc2)] - dz
        )
        # f = np.maximum(f, temp)  # bad performance in python3. Replaced with fmax
        f = np.fmax(f, temp)
        index += 1.0

    f = f - a
    f = np.logical_not(f)
    sh = np.double(f)

    return sh


# @jit(nopython=True)
def shadowingfunction_20(a, vegdem, vegdem2, azimuth, altitude, scale, amaxvalue, bush, forsvf):
    # plt.ion()
    # fig = plt.figure(figsize=(24, 7))
    # plt.axis('image')
    # ax1 = plt.subplot(2, 3, 1)
    # ax2 = plt.subplot(2, 3, 2)
    # ax3 = plt.subplot(2, 3, 3)
    # ax4 = plt.subplot(2, 3, 4)
    # ax5 = plt.subplot(2, 3, 5)
    # ax6 = plt.subplot(2, 3, 6)
    # ax1.title.set_text('fabovea')
    # ax2.title.set_text('gabovea')
    # ax3.title.set_text('vegsh at ' + str(altitude))
    # ax4.title.set_text('lastfabovea')
    # ax5.title.set_text('lastgabovea')
    # ax6.title.set_text('vegdem')

    # This function casts shadows on buildings and vegetation units.
    # New capability to deal with pergolas 20210827

    # conversion
    degrees = np.pi / 180.0
    azimuth = azimuth * degrees
    altitude = altitude * degrees

    # measure the size of grid
    sizex = a.shape[0]
    sizey = a.shape[1]

    # progressbar for svf plugin
    if forsvf == 0:
        barstep = np.max([sizex, sizey])
        total = 100.0 / barstep
        # dlg.progressBar.setRange(0, barstep)
        # dlg.progressBar.setValue(0)

    # initialise parameters
    dx = 0.0
    dy = 0.0
    dz = 0.0
    temp = np.zeros((sizex, sizey), dtype=np.float32)
    tempvegdem = np.zeros((sizex, sizey), dtype=np.float32)
    tempvegdem2 = np.zeros((sizex, sizey), dtype=np.float32)
    templastfabovea = np.zeros((sizex, sizey), dtype=np.float32)
    templastgabovea = np.zeros((sizex, sizey), dtype=np.float32)
    bushplant = bush > 1.0
    sh = np.zeros((sizex, sizey), dtype=np.float32)  # shadows from buildings
    vbshvegsh = np.zeros((sizex, sizey), dtype=np.float32)  # vegetation blocking buildings
    vegsh = np.add(np.zeros((sizex, sizey), dtype=np.float32), bushplant, dtype=float)  # vegetation shadow
    f = a

    pibyfour = np.pi / 4.0
    threetimespibyfour = 3.0 * pibyfour
    fivetimespibyfour = 5.0 * pibyfour
    seventimespibyfour = 7.0 * pibyfour
    sinazimuth = np.sin(azimuth)
    cosazimuth = np.cos(azimuth)
    tanazimuth = np.tan(azimuth)
    signsinazimuth = np.sign(sinazimuth)
    signcosazimuth = np.sign(cosazimuth)
    dssin = np.abs(1.0 / sinazimuth)
    dscos = np.abs(1.0 / cosazimuth)
    tanaltitudebyscale = np.tan(altitude) / scale
    # index = 1
    index = 0

    # new case with pergola (thin vertical layer of vegetation), August 2021
    dzprev = 0

    # main loop
    while (amaxvalue >= dz) and (np.abs(dx) < sizex) and (np.abs(dy) < sizey):
        if (
            (pibyfour <= azimuth)
            and (azimuth < threetimespibyfour)
            or (fivetimespibyfour <= azimuth)
            and (azimuth < seventimespibyfour)
        ):
            dy = signsinazimuth * index
            dx = -1.0 * signcosazimuth * np.abs(np.round(index / tanazimuth))
            ds = dssin
        else:
            dy = signsinazimuth * np.abs(np.round(index * tanazimuth))
            dx = -1.0 * signcosazimuth * index
            ds = dscos
        # note: dx and dy represent absolute values while ds is an incremental value
        dz = (ds * index) * tanaltitudebyscale
        tempvegdem[0:sizex, 0:sizey] = 0.0
        tempvegdem2[0:sizex, 0:sizey] = 0.0
        temp[0:sizex, 0:sizey] = 0.0
        templastfabovea[0:sizex, 0:sizey] = 0.0
        templastgabovea[0:sizex, 0:sizey] = 0.0
        absdx = np.abs(dx)
        absdy = np.abs(dy)
        xc1 = int((dx + absdx) / 2.0)
        xc2 = int(sizex + (dx - absdx) / 2.0)
        yc1 = int((dy + absdy) / 2.0)
        yc2 = int(sizey + (dy - absdy) / 2.0)
        xp1 = int(-((dx - absdx) / 2.0))
        xp2 = int(sizex - (dx + absdx) / 2.0)
        yp1 = int(-((dy - absdy) / 2.0))
        yp2 = int(sizey - (dy + absdy) / 2.0)

        tempvegdem[xp1:xp2, yp1:yp2] = vegdem[xc1:xc2, yc1:yc2] - dz
        tempvegdem2[xp1:xp2, yp1:yp2] = vegdem2[xc1:xc2, yc1:yc2] - dz
        temp[xp1:xp2, yp1:yp2] = a[xc1:xc2, yc1:yc2] - dz

        f = np.fmax(f, temp)  # Moving building shadow
        sh[(f > a)] = 1.0
        sh[(f <= a)] = 0.0
        fabovea = tempvegdem > a  # vegdem above DEM
        gabovea = tempvegdem2 > a  # vegdem2 above DEM

        # new pergola condition
        templastfabovea[xp1:xp2, yp1:yp2] = vegdem[xc1:xc2, yc1:yc2] - dzprev
        templastgabovea[xp1:xp2, yp1:yp2] = vegdem2[xc1:xc2, yc1:yc2] - dzprev
        lastfabovea = templastfabovea > a
        lastgabovea = templastgabovea > a
        dzprev = dz
        vegsh2 = np.add(
            np.add(np.add(fabovea, gabovea, dtype=float), lastfabovea, dtype=float),
            lastgabovea,
            dtype=float,
        )
        vegsh2[vegsh2 == 4] = 0.0
        # vegsh2[vegsh2 == 1] = 0. # This one is the ultimate question...
        vegsh2[vegsh2 > 0] = 1.0

        vegsh = np.fmax(vegsh, vegsh2)
        vegsh[(vegsh * sh > 0.0)] = 0.0
        vbshvegsh = vegsh + vbshvegsh  # removing shadows 'behind' buildings

        # im1 = ax1.imshow(fabovea)
        # im2 = ax2.imshow(gabovea)
        # im3 = ax3.imshow(vegsh)
        # im4 = ax4.imshow(lastfabovea)
        # im5 = ax5.imshow(lastgabovea)
        # im6 = ax6.imshow(vegshtest)
        # im1 = ax1.imshow(tempvegdem)
        # im2 = ax2.imshow(tempvegdem2)
        # im3 = ax3.imshow(vegsh)
        # im4 = ax4.imshow(templastfabovea)
        # im5 = ax5.imshow(templastgabovea)
        # im6 = ax6.imshow(vegshtest)
        # plt.show()
        # plt.pause(0.05)

        index += 1.0

    sh = 1.0 - sh
    vbshvegsh[(vbshvegsh > 0.0)] = 1.0
    vbshvegsh = vbshvegsh - vegsh
    vegsh = 1.0 - vegsh
    vbshvegsh = 1.0 - vbshvegsh

    # plt.close()
    # plt.ion()
    # fig = plt.figure(figsize=(24, 7))
    # plt.axis('image')
    # ax1 = plt.subplot(1, 3, 1)
    # im1 = ax1.imshow(vegsh)
    # plt.colorbar(im1)

    # ax2 = plt.subplot(1, 3, 2)
    # im2 = ax2.imshow(vegdem2)
    # plt.colorbar(im2)
    # plt.title('TDSM')

    # ax3 = plt.subplot(1, 3, 3)
    # im3 = ax3.imshow(vegdem)
    # plt.colorbar(im3)
    # plt.tight_layout()
    # plt.title('CDSM')
    # plt.show()
    # plt.pause(0.05)

    shadowresult = {"sh": sh, "vegsh": vegsh, "vbshvegsh": vbshvegsh}

    return shadowresult


# NOTE: Numba offers limited gains in this case
# @njit
def shadowingfunction_20_numba(
    a: np.ndarray,
    vegdem: np.ndarray,
    vegdem2: np.ndarray,
    azimuth: float,
    altitude: float,
    scale: float,
    amaxvalue: float,
    bush: np.ndarray,
    forsvf: int,
):  #  -> types.Tuple((types.float64[:, :], types.float64[:, :], types.float64[:, :])):
    # This function casts shadows on buildings and vegetation units.
    # New capability to deal with pergolas 20210827

    # conversion
    degrees = np.pi / 180.0
    azimuth = azimuth * degrees
    altitude = altitude * degrees

    # measure the size of grid
    sizex = a.shape[0]
    sizey = a.shape[1]

    # initialise parameters
    dx = 0.0
    dy = 0.0
    dz = 0.0
    temp = np.zeros((sizex, sizey))
    tempvegdem = np.zeros((sizex, sizey))
    tempvegdem2 = np.zeros((sizex, sizey))
    templastfabovea = np.zeros((sizex, sizey))
    templastgabovea = np.zeros((sizex, sizey))
    bushplant = bush > 1.0
    sh = np.zeros((sizex, sizey))  # shadows from buildings
    vbshvegsh = np.zeros((sizex, sizey))  # vegetation blocking buildings
    vegsh = np.zeros((sizex, sizey), dtype=np.float64)  # Initialize the array with zeros
    # Add bushplant values to the vegsh array
    for i in range(sizex):
        for j in range(sizey):
            vegsh[i, j] = bushplant[i, j]  # Assuming bushplant has the same shape
    f = a.copy()

    pibyfour = np.pi / 4.0
    threetimespibyfour = 3.0 * pibyfour
    fivetimespibyfour = 5.0 * pibyfour
    seventimespibyfour = 7.0 * pibyfour
    sinazimuth = np.sin(azimuth)
    cosazimuth = np.cos(azimuth)
    tanazimuth = np.tan(azimuth)
    signsinazimuth = np.sign(sinazimuth)
    signcosazimuth = np.sign(cosazimuth)
    dssin = np.abs(1.0 / sinazimuth) if sinazimuth != 0 else np.inf  # Avoid division by zero
    dscos = np.abs(1.0 / cosazimuth) if cosazimuth != 0 else np.inf  # Avoid division by zero
    tanaltitudebyscale = np.tan(altitude) / scale
    # index = 1
    index = 0

    # new case with pergola (thin vertical layer of vegetation), August 2021
    dzprev = 0

    # main loop
    while (amaxvalue >= dz) and (np.abs(dx) < sizex) and (np.abs(dy) < sizey):
        if (
            (pibyfour <= azimuth)
            and (azimuth < threetimespibyfour)
            or (fivetimespibyfour <= azimuth)
            and (azimuth < seventimespibyfour)
        ):
            dy = signsinazimuth * index
            dx = -1.0 * signcosazimuth * np.abs(np.round(index / tanazimuth))
            ds = dssin
        else:
            dy = signsinazimuth * np.abs(np.round(index * tanazimuth))
            dx = -1.0 * signcosazimuth * index
            ds = dscos
        # note: dx and dy represent absolute values while ds is an incremental value
        dz = (ds * index) * tanaltitudebyscale
        print(index, dz)
        tempvegdem[0:sizex, 0:sizey] = 0.0
        tempvegdem2[0:sizex, 0:sizey] = 0.0
        temp[0:sizex, 0:sizey] = 0.0
        templastfabovea[0:sizex, 0:sizey] = 0.0
        templastgabovea[0:sizex, 0:sizey] = 0.0
        absdx = np.abs(dx)
        absdy = np.abs(dy)
        xc1 = int((dx + absdx) / 2.0)
        xc2 = int(sizex + (dx - absdx) / 2.0)
        yc1 = int((dy + absdy) / 2.0)
        yc2 = int(sizey + (dy - absdy) / 2.0)
        xp1 = int(-((dx - absdx) / 2.0))
        xp2 = int(sizex - (dx + absdx) / 2.0)
        yp1 = int(-((dy - absdy) / 2.0))
        yp2 = int(sizey - (dy + absdy) / 2.0)

        tempvegdem[xp1:xp2, yp1:yp2] = vegdem[xc1:xc2, yc1:yc2] - dz
        tempvegdem2[xp1:xp2, yp1:yp2] = vegdem2[xc1:xc2, yc1:yc2] - dz
        temp[xp1:xp2, yp1:yp2] = a[xc1:xc2, yc1:yc2] - dz

        f = np.fmax(f, temp)  # Moving building shadow
        for i in range(f.shape[0]):
            for j in range(f.shape[1]):
                if f[i, j] > a[i, j]:
                    sh[i, j] = 1.0
                else:
                    sh[i, j] = 0.0
        fabovea = tempvegdem > a  # vegdem above DEM
        gabovea = tempvegdem2 > a  # vegdem2 above DEM

        # new pergola condition
        templastfabovea[xp1:xp2, yp1:yp2] = vegdem[xc1:xc2, yc1:yc2] - dzprev
        templastgabovea[xp1:xp2, yp1:yp2] = vegdem2[xc1:xc2, yc1:yc2] - dzprev
        lastfabovea = templastfabovea > a
        lastgabovea = templastgabovea > a
        dzprev = dz
        # Initialize vegsh2 with the same shape as the input arrays
        vegsh2 = np.zeros(fabovea.shape, dtype=np.float64)
        # Perform the additions directly
        vegsh2 += fabovea
        vegsh2 += gabovea
        vegsh2 += lastfabovea
        vegsh2 += lastgabovea

        for i in range(vegsh2.shape[0]):
            for j in range(vegsh2.shape[1]):
                if vegsh2[i, j] == 4:
                    vegsh2[i, j] = 0.0
                if vegsh2[i, j] > 0:
                    vegsh2[i, j] = 1.0

        vegsh = np.fmax(vegsh, vegsh2)
        for i in range(vegsh.shape[0]):
            for j in range(vegsh.shape[1]):
                if vegsh[i, j] * sh[i, j] > 0.0:
                    vegsh[i, j] = 0.0
        vbshvegsh = vegsh + vbshvegsh  # removing shadows 'behind' buildings

        index += 1.0

    sh = 1.0 - sh
    for i in range(vbshvegsh.shape[0]):
        for j in range(vbshvegsh.shape[1]):
            if vbshvegsh[i, j] > 0.0:
                vbshvegsh[i, j] = 1.0
    vbshvegsh = vbshvegsh - vegsh
    vegsh = 1.0 - vegsh
    vbshvegsh = 1.0 - vbshvegsh

    return sh, vegsh, vbshvegsh



# -*- coding: utf-8 -*-
"""
This code is a translation of radiation related function in the COMFA model

Translated on Fri Oct 23 
@author: xlinfr
"""

from radiationfunctionsCOMFA import CNRRabs_Total
from radiationfunctionsCOMFA import Rad_Total_solweig
import numpy as np


# Radiation components in w m-2
Lin = 200.
Lup = 300.
Kin = 800.
Kup = 200.

# Time variables used in solweig version
yyyy = 2020
month = 6
day = 23
hour = 12
minu = 0

# yyyy = 2020
# month = 12
# day = 20
# hour = 12
# minu = 0

# Time vaiables used in original
t=hour + minu/60. # decimal time of day (up to 24)
if (yyyy % 4) == 0:
    if (yyyy % 100) == 0:
        if (yyyy % 400) == 0:
            leapyear = 1
        else:
            leapyear = 0
    else:
        leapyear = 1
else:
    leapyear = 0
if leapyear == 1:
    dayspermonth = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
else:
    dayspermonth = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

doy = np.sum(dayspermonth[0:month - 1]) + day # day of year (Jan 1 = 1)

# Lubbock 33.5779,-101.8552

lat = 33.5779 # latitude
lon = -101.8552 # longitude
alt = 992. # Altitude (m asl)
utc = -6 # standard time lubbock

Atr=0.7 # atmospheric transmittance
alpha = 0.37 # albedo of the cylinder
emis = 0.95 # emissivity of the cylinder
L=0.1 # length of cylinder (cm)
D=0.01 # Diameter of cylinder (cm)

Aeff=0.78 # Effective area of body. 0.78 for standing from Campebll and Normal (1998) (0.70 for sitting)


Total_CNRRabs_solweig, zensolweig, Kdsolweig = Rad_Total_solweig(L,D,Lin,Lup,Kin,Kup,yyyy,month,day,hour,minu,lat,lon,alt, emis,alpha,Aeff,utc, Kd=None, Ta=None, RH=None)

Total_CNRRabs, zenCOMFA, kdCOMFA = CNRRabs_Total(alpha, L, D, Lin, Lup, Kin, Kup, Atr, doy, t, np.array([lat]), alt, emis, Aeff)  #np.array([lat]) to deal with timeseries?

print('Total_CNRRabs: ' + str(Total_CNRRabs))
print('altComfa: ' + str(90-zenCOMFA))
print('Total_CNRRabs_solweig: ' + str(Total_CNRRabs_solweig))
print('ZenSolweig: ' + str(90-zensolweig))



"""
/***************************************************************************
 ProcessingUMEP
                                 A QGIS plugin
 UMEP for processing toolbox
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2020-04-02
        copyright            : (C) 2020 by Fredrik Lindberg
        email                : fredrikl@gvc.gu.se
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

"The Sky View Factor algorithm can be used to generate pixel wise sky view factor (SVF) "
"using ground and building digital surface models (DSM). Optionally, vegetation DSMs could also be used. "
"By definition, SVF is the ratio of the radiation received (or emitted) by a planar surface to the "
"radiation emitted (or received) by the entire hemispheric environment (Watson and Johnson 1987). "
"It is a dimensionless measure between zero and one, representing totally obstructed and free spaces, "
"respectively. The methodology that is used to generate SVF here is described in Lindberg and Grimmond (2010).\n"
"-------------\n"
"Lindberg F, Grimmond CSB (2010) Continuous sky view factor maps from high resolution urban digital elevation models. Clim Res 42:177–183\n"
"Watson ID, Johnson GT (1987) Graphical estimation of skyview-factors in urban environments. J Climatol 7: 193–197"
"------------\n"
"Full manual available via the <b>Help</b>-button."

"https://umep-docs.readthedocs.io/en/latest/pre-processor/Urban%20Geometry%20Sky%20View%20Factor%20Calculator.html"
"""

__author__ = "Fredrik Lindberg"
__date__ = "2020-04-02"
__copyright__ = "(C) 2020 by Fredrik Lindberg"

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = "$Format:%H$"

# %%
import logging
import os
import zipfile
from pathlib import Path

import numpy as np

from umep import class_configs, common
from umep.functions import svf_functions as svf

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


# %%
def generate_svf(
    dsm_path: str,
    bbox: list[int],
    out_dir: str,
    dem_path: str | None = None,
    cdsm_path: str | None = None,
    trans_veg_perc: float = 3,
    trunk_ratio_perc: float = 25,
    amax_local_window_m: int = 100,
    amax_local_perc: float = 99.9,
):
    out_path = Path(out_dir)
    out_path.mkdir(parents=True, exist_ok=True)
    out_path_str = str(out_path)

    # Open the DSM file
    dsm, dsm_trf, dsm_crs, dsm_nd = common.load_raster(dsm_path, bbox)
    dsm_pix_size = dsm_trf[1]
    dsm_scale = 1 / dsm_pix_size

    dem = None
    if dem_path is not None:
        dem, dem_trf, dem_crs, _dem_nd = common.load_raster(dem_path, bbox)
        assert dem.shape == dsm.shape, "Mismatching raster shapes for DSM and DEM."
        assert np.allclose(dsm_trf, dem_trf), "Mismatching spatial transform for DSM and DEM."
        assert dem_crs == dsm_crs, "Mismatching CRS for DSM and DEM."

    use_cdsm = False
    cdsm = None
    if cdsm_path is not None:
        use_cdsm = True
        cdsm, cdsm_trf, cdsm_crs, _cdsm_nd = common.load_raster(cdsm_path, bbox)
        assert cdsm.shape == dsm.shape, "Mismatching raster shapes for DSM and CDSM."
        assert np.allclose(dsm_trf, cdsm_trf), "Mismatching spatial transform for DSM and CDSM."
        assert cdsm_crs == dsm_crs, "Mismatching CRS for DSM and CDSM."

    # veg transmissivity as percentage
    if not (0 <= trans_veg_perc <= 100):
        raise ValueError("Vegetation transmissivity should be a number between 0 and 100")

    trans_veg = trans_veg_perc / 100.0
    trunk_ratio = trunk_ratio_perc / 100.0

    dsm, dem, cdsm, tdsm, amax = class_configs.raster_preprocessing(
        dsm,
        dem,
        cdsm,
        None,
        trunk_ratio,
        dsm_pix_size,
        amax_local_window_m=amax_local_window_m,
        amax_local_perc=amax_local_perc,
    )

    common.save_raster(
        out_path_str + "/input-dsm.tif",
        dsm,
        dsm_trf,
        dsm_crs,
        dsm_nd,
    )
    if dem is not None:
        common.save_raster(
            out_path_str + "/input-dem.tif",
            dem,
            dsm_trf,
            dsm_crs,
            dsm_nd,
        )
    if use_cdsm:
        common.save_raster(
            out_path_str + "/input-cdsm.tif",
            cdsm,
            dsm_trf,
            dsm_crs,
            dsm_nd,
        )
        common.save_raster(
            out_path_str + "/input-tdsm.tif",
            tdsm,
            dsm_trf,
            dsm_crs,
            dsm_nd,
        )

    # compute
    ret = svf.svfForProcessing153(dsm, cdsm, tdsm, dsm_scale, use_cdsm, amax)

    svfbu = ret["svf"]
    svfbuE = ret["svfE"]
    svfbuS = ret["svfS"]
    svfbuW = ret["svfW"]
    svfbuN = ret["svfN"]

    # Save the rasters using rasterio
    common.save_raster(out_path_str + "/" + "svf.tif", svfbu, dsm_trf, dsm_crs)
    common.save_raster(out_path_str + "/" + "svfE.tif", svfbuE, dsm_trf, dsm_crs)
    common.save_raster(out_path_str + "/" + "svfS.tif", svfbuS, dsm_trf, dsm_crs)
    common.save_raster(out_path_str + "/" + "svfW.tif", svfbuW, dsm_trf, dsm_crs)
    common.save_raster(out_path_str + "/" + "svfN.tif", svfbuN, dsm_trf, dsm_crs)

    # Create or update the ZIP file
    zip_filepath = out_path_str + "/" + "svfs.zip"
    if os.path.isfile(zip_filepath):
        os.remove(zip_filepath)

    with zipfile.ZipFile(zip_filepath, "a") as zippo:
        zippo.write(out_path_str + "/" + "svf.tif", "svf.tif")
        zippo.write(out_path_str + "/" + "svfE.tif", "svfE.tif")
        zippo.write(out_path_str + "/" + "svfS.tif", "svfS.tif")
        zippo.write(out_path_str + "/" + "svfW.tif", "svfW.tif")
        zippo.write(out_path_str + "/" + "svfN.tif", "svfN.tif")

    # Remove the individual TIFF files after zipping
    os.remove(out_path_str + "/" + "svf.tif")
    os.remove(out_path_str + "/" + "svfE.tif")
    os.remove(out_path_str + "/" + "svfS.tif")
    os.remove(out_path_str + "/" + "svfW.tif")
    os.remove(out_path_str + "/" + "svfN.tif")

    if use_cdsm == 0:
        svftotal = svfbu
    else:
        # Report the vegetation-related results
        svfveg = ret["svfveg"]
        svfEveg = ret["svfEveg"]
        svfSveg = ret["svfSveg"]
        svfWveg = ret["svfWveg"]
        svfNveg = ret["svfNveg"]
        svfaveg = ret["svfaveg"]
        svfEaveg = ret["svfEaveg"]
        svfSaveg = ret["svfSaveg"]
        svfWaveg = ret["svfWaveg"]
        svfNaveg = ret["svfNaveg"]

        # Save vegetation rasters
        common.save_raster(out_path_str + "/" + "svfveg.tif", svfveg, dsm_trf, dsm_crs)
        common.save_raster(out_path_str + "/" + "svfEveg.tif", svfEveg, dsm_trf, dsm_crs)
        common.save_raster(out_path_str + "/" + "svfSveg.tif", svfSveg, dsm_trf, dsm_crs)
        common.save_raster(out_path_str + "/" + "svfWveg.tif", svfWveg, dsm_trf, dsm_crs)
        common.save_raster(out_path_str + "/" + "svfNveg.tif", svfNveg, dsm_trf, dsm_crs)
        common.save_raster(out_path_str + "/" + "svfaveg.tif", svfaveg, dsm_trf, dsm_crs)
        common.save_raster(out_path_str + "/" + "svfEaveg.tif", svfEaveg, dsm_trf, dsm_crs)
        common.save_raster(out_path_str + "/" + "svfSaveg.tif", svfSaveg, dsm_trf, dsm_crs)
        common.save_raster(out_path_str + "/" + "svfWaveg.tif", svfWaveg, dsm_trf, dsm_crs)
        common.save_raster(out_path_str + "/" + "svfNaveg.tif", svfNaveg, dsm_trf, dsm_crs)

        # Add vegetation rasters to the ZIP file
        with zipfile.ZipFile(zip_filepath, "a") as zippo:
            zippo.write(out_path_str + "/" + "svfveg.tif", "svfveg.tif")
            zippo.write(out_path_str + "/" + "svfEveg.tif", "svfEveg.tif")
            zippo.write(out_path_str + "/" + "svfSveg.tif", "svfSveg.tif")
            zippo.write(out_path_str + "/" + "svfWveg.tif", "svfWveg.tif")
            zippo.write(out_path_str + "/" + "svfNveg.tif", "svfNveg.tif")
            zippo.write(out_path_str + "/" + "svfaveg.tif", "svfaveg.tif")
            zippo.write(out_path_str + "/" + "svfEaveg.tif", "svfEaveg.tif")
            zippo.write(out_path_str + "/" + "svfSaveg.tif", "svfSaveg.tif")
            zippo.write(out_path_str + "/" + "svfWaveg.tif", "svfWaveg.tif")
            zippo.write(out_path_str + "/" + "svfNaveg.tif", "svfNaveg.tif")

        # Remove the individual TIFF files after zipping
        os.remove(out_path_str + "/" + "svfveg.tif")
        os.remove(out_path_str + "/" + "svfEveg.tif")
        os.remove(out_path_str + "/" + "svfSveg.tif")
        os.remove(out_path_str + "/" + "svfWveg.tif")
        os.remove(out_path_str + "/" + "svfNveg.tif")
        os.remove(out_path_str + "/" + "svfaveg.tif")
        os.remove(out_path_str + "/" + "svfEaveg.tif")
        os.remove(out_path_str + "/" + "svfSaveg.tif")
        os.remove(out_path_str + "/" + "svfWaveg.tif")
        os.remove(out_path_str + "/" + "svfNaveg.tif")

        # Calculate final total SVF
        svftotal = svfbu - (1 - svfveg) * (1 - trans_veg)

    # Save the final svftotal raster
    common.save_raster(out_path_str + "/" + "svf_total.tif", svftotal, dsm_trf, dsm_crs)

    # Save shadow matrices as compressed npz
    shmat = ret["shmat"]
    vegshmat = ret["vegshmat"]
    vbshvegshmat = ret["vbshvegshmat"]

    np.savez_compressed(
        out_path_str + "/" + "shadowmats.npz",
        shadowmat=shmat,
        vegshadowmat=vegshmat,
        vbshmat=vbshvegshmat,
    )



import numpy as np

''' Model 1 is based on Unsworth & Monteith, 1975 '''
def model1(sky_patches, esky, Ta):

    # Stefan-Boltzmann's Constant
    SBC = 5.67051e-8

    # Degrees to radians
    deg2rad = np.pi / 180               

    # Unique altitudes in lv, i.e. unique altitude for the patches
    skyalt, skyalt_c = np.unique(sky_patches[:, 0], return_counts=True)
    
    # Unique zeniths for the patches
    skyzen = 90-skyalt                  
    
    # Cosine of the zenith angles
    cosskyzen = np.cos(skyzen * deg2rad)

    # Estimate emissivities at different altitudes/zenith angles
    
    # Constants?
    a_c = 0.67
    b_c = 0.094

    # Natural log of the reduced depth of precipitable water
    ln_u_prec = esky/b_c-a_c/b_c-0.5
    
    # Reduced depth of precipitable water    
    u_prec = np.exp(ln_u_prec)          

    # Optical water depth
    owp = u_prec/cosskyzen              

    # Natural log of optical water depth
    log_owp = np.log(owp)               

    # Emissivity of each zenith angle, i.e. the zenith angle of each band of patches
    esky_band = a_c+b_c*log_owp            
    
    # Altitudes of the Robinson & Stone patches
    p_alt = sky_patches[:,0]
    
    # Empty array with length based on number of patches                
    patch_emissivity = np.zeros((p_alt.shape[0]))

    # Fill with emissivities
    for idx in skyalt:
        temp_emissivity = esky_band[skyalt == idx]
        patch_emissivity[p_alt == idx] = temp_emissivity

    # Normalize
    patch_emissivity_normalized = patch_emissivity/np.sum(patch_emissivity)

    return patch_emissivity_normalized, esky_band

''' Model 2 is based on Martin & Berdhal, 1984
    Ez = 1 - (1 - Es)*e**(b2*(1.7 - (1/np.cos(z)))) '''
def model2(sky_patches, esky, Ta):

    # Degrees to radians
    deg2rad = np.pi / 180               

    # Unique altitudes in lv, i.e. unique altitude for the patches
    skyalt, skyalt_c = np.unique(sky_patches[:, 0], return_counts=True)
    
    # Unique zeniths for the patches
    skyzen = 90-skyalt                  
    
    # Constant (Ångström, 1915), proposed by Nahon et al., 2019
    b_c = 0.308
    # b_c = 0.1

    # Estimate emissivites at different altitudes/zenith angles
    esky_band = 1 - (1 - esky) * np.exp(b_c * (1.7 - (1 / np.cos(skyzen * deg2rad))))
    
    # Altitudes of the Robinson & Stone patches
    p_alt = sky_patches[:,0]
    
    # Empty array with length based on number of patches                
    patch_emissivity = np.zeros((p_alt.shape[0]))

    # Fill with emissivities
    for idx in skyalt:
        temp_emissivity = esky_band[skyalt == idx]
        patch_emissivity[p_alt == idx] = temp_emissivity

    # Normalize
    patch_emissivity_normalized = patch_emissivity/np.sum(patch_emissivity)

    return patch_emissivity_normalized, esky_band

''' Model 3 is based on Bliss, 1961.
    Ez = 1 - (1 - Es)**(1/(b1*np.cos(z))) '''
def model3(sky_patches, esky, Ta):

    # Degrees to radians
    deg2rad = np.pi / 180               

    # Unique altitudes in lv, i.e. unique altitude for the patches
    skyalt, skyalt_c = np.unique(sky_patches[:, 0], return_counts=True)
    
    # Unique zeniths for the patches
    skyzen = 90-skyalt                  
    
    # Constant, can (should?) be changed. Model gave unsatisfactory results in Nahon et al., 2019
    b_c = 1.8

    # Estimate emissivites at different altitudes/zenith angles
    esky_band = 1 - (1 - esky)**(1/(b_c * np.cos(skyzen * deg2rad)))

    # Estimating longwave radiation for each patch to a horizontal or vertical surface
    
    # Altitudes of the Robinson & Stone patches
    p_alt = sky_patches[:,0]
    
    # Empty array with length based on number of patches                
    patch_emissivity = np.zeros((p_alt.shape[0]))

    # Fill with emissivities
    for idx in skyalt:
        temp_emissivity = esky_band[skyalt == idx]
        patch_emissivity[p_alt == idx] = temp_emissivity

    # Normalize
    patch_emissivity_normalized = patch_emissivity/np.sum(patch_emissivity)

    return patch_emissivity_normalized, esky_band


import numpy as np

def create_patches(patch_option):

    deg2rad = np.pi/180

    # patch_option = 1 = 145 patches (Robinson & Stone, 2004)
    # patch_option = 2 = 153 patches (Wallenberg et al., 2022)
    # patch_option = 3 = 306 patches -> test 
    # patch_option = 4 = 612 patches -> test

    skyvaultalt = np.atleast_2d([])
    skyvaultazi = np.atleast_2d([])
            
    # Creating skyvault of patches of constant radians (Tregeneza and Sharples, 1993)
    # Patch option 1, 145 patches, Original Robinson & Stone (2004) after Tregenza (1987)/Tregenza & Sharples (1993)
    if patch_option == 1:
        annulino = np.array([0, 12, 24, 36, 48, 60, 72, 84, 90])
        skyvaultaltint = np.array([6, 18, 30, 42, 54, 66, 78, 90]) # Robinson & Stone (2004)
        azistart = np.array([0, 4, 2, 5, 8, 0, 10, 0]) # Fredrik/Nils
        patches_in_band = np.array([30, 30, 24, 24, 18, 12, 6, 1]) # Robinson & Stone (2004)
    # Patch option 2, 153 patches, Wallenberg et al. (2022)
    elif patch_option == 2:
        annulino = np.array([0, 12, 24, 36, 48, 60, 72, 84, 90])
        skyvaultaltint = np.array([6, 18, 30, 42, 54, 66, 78, 90]) # Robinson & Stone (2004)
        azistart = np.array([0, 4, 2, 5, 8, 0, 10, 0]) # Fredrik/Nils
        patches_in_band = np.array([31, 30, 28, 24, 19, 13, 7, 1]) # Nils
    # Patch option 3, 306 patches, test
    elif patch_option == 3:
        annulino = np.array([0, 12, 24, 36, 48, 60, 72, 84, 90])
        skyvaultaltint = np.array([6, 18, 30, 42, 54, 66, 78, 90]) # Robinson & Stone (2004)
        azistart = np.array([0, 4, 2, 5, 8, 0, 10, 0]) # Fredrik/Nils
        patches_in_band = np.array([31*2, 30*2, 28*2, 24*2, 19*2, 13*2, 7*2, 1]) # Nils
    # Patch option 4, 612 patches, test
    elif patch_option == 4:
        annulino = np.array([0, 4.5, 9, 15, 21, 27, 33, 39, 45, 51, 57, 63, 69, 75, 81, 90]) # Nils
        skyvaultaltint = np.array([3, 9, 15, 21, 27, 33, 39, 45, 51, 57, 63, 69, 75, 81, 90]) # Nils
        patches_in_band = np.array([31*2, 31*2, 30*2, 30*2, 28*2, 28*2, 24*2, 24*2, 19*2, 19*2, 13*2, 13*2, 7*2, 7*2, 1]) # Nils
        azistart = np.array([0, 0, 4, 4, 2, 2, 5, 5, 8, 8, 0, 0, 10, 10, 0]) # Nils

    skyvaultaziint = np.array([360/patches for patches in patches_in_band])

    for j in range(0, skyvaultaltint.shape[0]):
        for k in range(0, patches_in_band[j]):
            skyvaultalt = np.append(skyvaultalt, skyvaultaltint[j])
            skyvaultazi = np.append(skyvaultazi, k*skyvaultaziint[j] + azistart[j])

    # skyvaultzen = (90 - skyvaultalt) * deg2rad
    # skyvaultalt = skyvaultalt * deg2rad
    # skyvaultazi = skyvaultazi * deg2rad

    return skyvaultalt, skyvaultazi, annulino, skyvaultaltint, patches_in_band, skyvaultaziint, azistart


import numpy as np

def sunonsurface_2018a(azimuthA, scale, buildings, shadow, sunwall, first, second, aspect, walls, Tg, Tgwall, Ta,
                       emis_grid, ewall, alb_grid, SBC, albedo_b, Twater, lc_grid, landcover):
    # This version of sunonsurfce goes with SOLWEIG 2015a. It also simulates
    # Lup and albedo based on landcover information and shadow patterns.
    # Fredrik Lindberg, fredrikl@gvc.gu.se

    sizex = np.shape(walls)[0]
    sizey = np.shape(walls)[1]

    # sizex=size(buildings,1);sizey=size(buildings,2);
    wallbol = (walls > 0) * 1
    sunwall[sunwall > 0] = 1  # test 20160910

    # conversion into radians
    azimuth = azimuthA * (np.pi / 180)

    # loop parameters
    index = 0
    f = buildings
    Lup = SBC * emis_grid * (Tg * shadow + Ta + 273.15) ** 4 - SBC * emis_grid * (Ta + 273.15) ** 4  # +Ta
    if landcover == 1:
        Tg[lc_grid == 3] = Twater - Ta  # Setting water temperature

    Lwall = SBC * ewall * (Tgwall + Ta + 273.15) ** 4 - SBC * ewall * (Ta + 273.15) ** 4  # +Ta
    albshadow = alb_grid * shadow
    alb = alb_grid
    # sh(sh<=0.1)=0;
    # sh=sh-(1-vegsh)*(1-psi);
    # shadow=sh-(1-vegsh)*(1-psi);
    # dx=0;
    # dy=0;
    # ds=0; ##ok<NASGU>

    tempsh = np.zeros((sizex, sizey))
    tempbu = np.zeros((sizex, sizey))
    tempbub = np.zeros((sizex, sizey))
    tempbubwall = np.zeros((sizex, sizey))
    tempwallsun = np.zeros((sizex, sizey))
    weightsumsh = np.zeros((sizex, sizey))
    weightsumwall = np.zeros((sizex, sizey))
    first = np.round(first * scale)
    if first < 1:
        first = 1
    second = np.round(second * scale)
    # tempTgsh=tempsh;
    weightsumLupsh = np.zeros((sizex, sizey))
    weightsumLwall = np.zeros((sizex, sizey))
    weightsumalbsh = np.zeros((sizex, sizey))
    weightsumalbwall = np.zeros((sizex, sizey))
    weightsumalbnosh = np.zeros((sizex, sizey))
    weightsumalbwallnosh = np.zeros((sizex, sizey))
    tempLupsh = np.zeros((sizex, sizey))
    tempalbsh = np.zeros((sizex, sizey))
    tempalbnosh = np.zeros((sizex, sizey))

    # other loop parameters
    pibyfour = np.pi / 4
    threetimespibyfour = 3 * pibyfour
    fivetimespibyfour = 5 * pibyfour
    seventimespibyfour = 7 * pibyfour
    sinazimuth = np.sin(azimuth)
    cosazimuth = np.cos(azimuth)
    tanazimuth = np.tan(azimuth)
    signsinazimuth = np.sign(sinazimuth)
    signcosazimuth = np.sign(cosazimuth)

    ## The Shadow casting algoritm
    for n in np.arange(0, second):
        if (pibyfour <= azimuth and azimuth < threetimespibyfour) or (
                fivetimespibyfour <= azimuth and azimuth < seventimespibyfour):
            dy = signsinazimuth * index
            dx = -1 * signcosazimuth * np.abs(np.round(index / tanazimuth))
            # ds = dssin
        else:
            dy = signsinazimuth * abs(round(index * tanazimuth))
            dx = -1 * signcosazimuth * index
            # ds = dscos

        absdx = np.abs(dx)
        absdy = np.abs(dy)

        xc1 = ((dx + absdx) / 2)
        xc2 = (sizex + (dx - absdx) / 2)
        yc1 = ((dy + absdy) / 2)
        yc2 = (sizey + (dy - absdy) / 2)

        xp1 = -((dx - absdx) / 2)
        xp2 = (sizex - (dx + absdx) / 2)
        yp1 = -((dy - absdy) / 2)
        yp2 = (sizey - (dy + absdy) / 2)

        tempbu[int(xp1):int(xp2), int(yp1):int(yp2)] = buildings[int(xc1):int(xc2),
                                                       int(yc1):int(yc2)]  # moving building
        tempsh[int(xp1):int(xp2), int(yp1):int(yp2)] = shadow[int(xc1):int(xc2), int(yc1):int(yc2)]  # moving shadow
        tempLupsh[int(xp1):int(xp2), int(yp1):int(yp2)] = Lup[int(xc1):int(xc2), int(yc1):int(yc2)]  # moving Lup/shadow
        tempalbsh[int(xp1):int(xp2), int(yp1):int(yp2)] = albshadow[int(xc1):int(xc2),
                                                          int(yc1):int(yc2)]  # moving Albedo/shadow
        tempalbnosh[int(xp1):int(xp2), int(yp1):int(yp2)] = alb[int(xc1):int(xc2), int(yc1):int(yc2)]  # moving Albedo
        f = np.min([f, tempbu], axis=0)  # utsmetning av buildings

        shadow2 = tempsh * f
        weightsumsh = weightsumsh + shadow2

        Lupsh = tempLupsh * f
        weightsumLupsh = weightsumLupsh + Lupsh

        albsh = tempalbsh * f
        weightsumalbsh = weightsumalbsh + albsh

        albnosh = tempalbnosh * f
        weightsumalbnosh = weightsumalbnosh + albnosh

        tempwallsun[int(xp1):int(xp2), int(yp1):int(yp2)] = sunwall[int(xc1):int(xc2),
                                                            int(yc1):int(yc2)]  # moving buildingwall insun image
        tempb = tempwallsun * f
        tempbwall = f * -1 + 1
        tempbub = ((tempb + tempbub) > 0) * 1
        tempbubwall = ((tempbwall + tempbubwall) > 0) * 1
        weightsumLwall = weightsumLwall + tempbub * Lwall
        weightsumalbwall = weightsumalbwall + tempbub * albedo_b
        weightsumwall = weightsumwall + tempbub
        weightsumalbwallnosh = weightsumalbwallnosh + tempbubwall * albedo_b

        ind = 1
        if (n + 1) <= first:
            weightsumwall_first = weightsumwall / ind
            weightsumsh_first = weightsumsh / ind
            wallsuninfluence_first = weightsumwall_first > 0
            weightsumLwall_first = (weightsumLwall) / ind  # *Lwall
            weightsumLupsh_first = weightsumLupsh / ind

            weightsumalbwall_first = weightsumalbwall / ind  # *albedo_b
            weightsumalbsh_first = weightsumalbsh / ind
            weightsumalbwallnosh_first = weightsumalbwallnosh / ind  # *albedo_b
            weightsumalbnosh_first = weightsumalbnosh / ind
            wallinfluence_first = weightsumalbwallnosh_first > 0
            #         gvf1=(weightsumwall+weightsumsh)/first;
            #         gvf1(gvf1>1)=1;
            ind += 1
        index += 1

    wallsuninfluence_second = weightsumwall > 0
    wallinfluence_second = weightsumalbwallnosh > 0
    # gvf2(gvf2>1)=1;

    # Removing walls in shadow due to selfshadowing
    azilow = azimuth - np.pi / 2
    azihigh = azimuth + np.pi / 2
    if azilow >= 0 and azihigh < 2 * np.pi:  # 90 to 270  (SHADOW)
        facesh = (np.logical_or(aspect < azilow, aspect >= azihigh).astype(float) - wallbol + 1)
    elif azilow < 0 and azihigh <= 2 * np.pi:  # 0 to 90
        azilow = azilow + 2 * np.pi
        facesh = np.logical_or(aspect > azilow, aspect <= azihigh) * -1 + 1  # (SHADOW)    # check for the -1
    elif azilow > 0 and azihigh >= 2 * np.pi:  # 270 to 360
        azihigh = azihigh - 2 * np.pi
        facesh = np.logical_or(aspect > azilow, aspect <= azihigh) * -1 + 1  # (SHADOW)

    # removing walls in self shadoing
    keep = (weightsumwall == second) - facesh
    keep[keep == -1] = 0

    # gvf from shadow only
    gvf1 = ((weightsumwall_first + weightsumsh_first) / (first + 1)) * wallsuninfluence_first + \
           (weightsumsh_first) / (first) * (wallsuninfluence_first * -1 + 1)
    weightsumwall[keep == 1] = 0
    gvf2 = ((weightsumwall + weightsumsh) / (second + 1)) * wallsuninfluence_second + \
           (weightsumsh) / (second) * (wallsuninfluence_second * -1 + 1)

    gvf2[gvf2 > 1.] = 1.

    # gvf from shadow and Lup
    gvfLup1 = ((weightsumLwall_first + weightsumLupsh_first) / (first + 1)) * wallsuninfluence_first + \
              (weightsumLupsh_first) / (first) * (wallsuninfluence_first * -1 + 1)
    weightsumLwall[keep == 1] = 0
    gvfLup2 = ((weightsumLwall + weightsumLupsh) / (second + 1)) * wallsuninfluence_second + \
              (weightsumLupsh) / (second) * (wallsuninfluence_second * -1 + 1)

    # gvf from shadow and albedo
    gvfalb1 = ((weightsumalbwall_first + weightsumalbsh_first) / (first + 1)) * wallsuninfluence_first + \
              (weightsumalbsh_first) / (first) * (wallsuninfluence_first * -1 + 1)
    weightsumalbwall[keep == 1] = 0
    gvfalb2 = ((weightsumalbwall + weightsumalbsh) / (second + 1)) * wallsuninfluence_second + \
              (weightsumalbsh) / (second) * (wallsuninfluence_second * -1 + 1)

    # gvf from albedo only
    gvfalbnosh1 = ((weightsumalbwallnosh_first + weightsumalbnosh_first) / (first + 1)) * wallinfluence_first + \
                  (weightsumalbnosh_first) / (first) * (wallinfluence_first * -1 + 1)  #
    gvfalbnosh2 = ((weightsumalbwallnosh + weightsumalbnosh) / (second)) * wallinfluence_second + \
                  (weightsumalbnosh) / (second) * (wallinfluence_second * -1 + 1)

    # Weighting
    gvf = (gvf1 * 0.5 + gvf2 * 0.4) / 0.9
    gvfLup = (gvfLup1 * 0.5 + gvfLup2 * 0.4) / 0.9
    gvfLup = gvfLup + ((SBC * emis_grid * (Tg * shadow + Ta + 273.15) ** 4) - SBC * emis_grid * (Ta + 273.15) ** 4) * (
                buildings * -1 + 1)  # +Ta
    gvfalb = (gvfalb1 * 0.5 + gvfalb2 * 0.4) / 0.9
    gvfalb = gvfalb + alb_grid * (buildings * -1 + 1) * shadow
    gvfalbnosh = (gvfalbnosh1 * 0.5 + gvfalbnosh2 * 0.4) / 0.9
    gvfalbnosh = gvfalbnosh * buildings + alb_grid * (buildings * -1 + 1)

    return gvf, gvfLup, gvfalb, gvfalbnosh, gvf2


# Generated with SMOP  0.41
# from libsmop import *
# CNRRabs_Total.m

    
# @function
def CNRRabs_Total(alpha=None,L=None,D=None,Lin=None,Lup=None,Kin=None,Kup=None,Atr=None,d=None,t=None,lat=None,A=None,*args,**kwargs):
    varargin = CNRRabs_Total.varargin
    nargin = CNRRabs_Total.nargin

    # VERTICAL CYLINDER MODEL 
    # Used to calculate the total radiation absorbed by a cylinder, such as the CRT, with inputs
    # alpha (albedo of cylinder), Kin (incoming shortwave measured by CNR net
    # radiometer, W/m2), Kup (reflected shortwave measured by CNR, W/m2), Lin
    # (CNR incoming longwave radiation, W/m2), Lup (CNR1 outgoing longwave
    # radition, W/m2), L (length of cylinder, m), D (diamter of cylinder, m).
    # 
    #Also need solar zenith, which is calculated from: lat (latitude, degrees), d (days, Jan 1=1), t (decimal
    # time, 24h decimal time).
        
    #Atr - atmospheric transmittance - can calculate for the day/time of day or
    #estimate based on sky conditions.
        
    # LOAD data ....
    
    #set constants
    D=0.01
    # CNRRabs_Total.m:20
    
    L=0.1
    # CNRRabs_Total.m:21
    
    alpha=0.37
    # CNRRabs_Total.m:22
    
    #     E.g., 
    #     Kin = metdata(:,10); #put in column for Kin from metdata file
    #     Kup = metdata(:,11); 
    #     Lin  = metdata(:,20);
    #     Lup = metdata(:,21); 
    #     d = metdata(:,24);
    #     Atr = 0.70; 
    #     lat = 33.75;
    #     A = 992;
    
    Aeff=0.78
    # CNRRabs_Total.m:33
    
    Kin_abs=CNR_Kinabs_meas(alpha,Kin,L,D,A,lat,d,t,Atr)
    # CNRRabs_Total.m:35
    Kup_abs=CNR_Kup(alpha,Kup,L,D)
    # CNRRabs_Total.m:37
    Lin_abs=LinMeas_abs(L,D,Lin)
    # CNRRabs_Total.m:39
    Lup_abs=LupMeas_abs(L,D,Lup)
    # CNRRabs_Total.m:41
    Acyl=CRT_Acyl(L,D)
    # CNRRabs_Total.m:43
    Total_CNRRabs=multiply(Aeff,((Kin_abs + Kup_abs + Lin_abs + Lup_abs) / (Acyl)))
    # CNRRabs_Total.m:46


import logging
from pathlib import Path

import numpy as np

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

try:
    import pyproj
    import rasterio
    from rasterio.features import rasterize
    from rasterio.mask import mask
    from rasterio.transform import Affine, from_origin
    from shapely import geometry

    GDAL_ENV = False
    logger.info("Using rasterio for raster operations.")

except:
    from osgeo import gdal

    GDAL_ENV = True
    logger.info("Using GDAL for raster operations.")


def rasterise_gdf(gdf, geom_col, ht_col, bbox=None, pixel_size: int = 1):
    # Define raster parameters
    if bbox is not None:
        # Unpack bbox values
        minx, miny, maxx, maxy = bbox
    else:
        # Use the total bounds of the GeoDataFrame
        minx, miny, maxx, maxy = gdf.total_bounds
    width = int((maxx - minx) / pixel_size)
    height = int((maxy - miny) / pixel_size)
    transform = from_origin(minx, maxy, pixel_size, pixel_size)
    # Create a blank array for the raster
    raster = np.zeros((height, width), dtype=np.float32)
    # Burn geometries into the raster
    shapes = ((geom, value) for geom, value in zip(gdf[geom_col], gdf[ht_col], strict=True))
    raster = rasterize(shapes, out_shape=raster.shape, transform=transform, fill=0, dtype=np.float32)

    return raster, transform


def check_path(path_str: str | Path, make_dir: bool = False) -> Path:
    # Ensure path exists
    path = Path(path_str).absolute()
    if not path.parent.exists():
        if make_dir:
            path.parent.mkdir(parents=True, exist_ok=True)
        else:
            raise OSError(f"Parent directory {path} does not exist. Set make_dir=True to create it.")
    if not path.exists() and not path.suffix:
        if make_dir:
            path.mkdir(parents=True, exist_ok=True)
        else:
            raise OSError(f"Path {path} does not exist. Set make_dir=True to create it.")
    return path


def save_raster(
    out_path_str: str, data_arr: np.ndarray, trf_arr: list[float], crs_wkt: str, no_data_val: float = -9999
):
    attempts = 2
    while attempts > 0:
        attempts -= 1
        try:
            # Save raster using GDAL or rasterio
            out_path = check_path(out_path_str, make_dir=True)
            height, width = data_arr.shape
            if GDAL_ENV is False:
                trf = Affine.from_gdal(*trf_arr)
                crs = None
                if crs_wkt:
                    crs = pyproj.CRS(crs_wkt)
                with rasterio.open(
                    out_path,
                    "w",
                    driver="GTiff",
                    height=height,
                    width=width,
                    count=1,
                    dtype=data_arr.dtype,
                    crs=crs,
                    transform=trf,
                    nodata=no_data_val,
                ) as dst:
                    dst.write(data_arr, 1)
            else:
                # Map numpy dtype to GDAL type
                dtype_map = {
                    np.dtype("uint8"): gdal.GDT_Byte,
                    np.dtype("int16"): gdal.GDT_Int16,
                    np.dtype("uint16"): gdal.GDT_UInt16,
                    np.dtype("int32"): gdal.GDT_Int32,
                    np.dtype("uint32"): gdal.GDT_UInt32,
                    np.dtype("float32"): gdal.GDT_Float32,
                    np.dtype("float64"): gdal.GDT_Float64,
                }
                gdal_dtype = dtype_map.get(data_arr.dtype, gdal.GDT_Float32)
                driver = gdal.GetDriverByName("GTiff")
                ds = driver.Create(str(out_path), width, height, 1, gdal_dtype)
                # trf is a list: [top left x, w-e pixel size, rotation, top left y, rotation, n-s pixel size]
                ds.SetGeoTransform(tuple(trf_arr))
                # GetProjection returns WKT (string)
                if crs_wkt:
                    ds.SetProjection(crs_wkt)
                band = ds.GetRasterBand(1)
                band.WriteArray(data_arr, 0, 0)
                band.SetNoDataValue(no_data_val)
                ds.FlushCache()
                ds = None
        except Exception as e:
            print(f"Error saving raster, attempts left {attempts}: {e}")
            if attempts == 0:
                raise e


def load_raster(
    path_str: str, bbox: list[int] | None = None, band: int = 0
) -> tuple[np.ndarray, list[float], str | None, float | None]:
    # Load raster, optionally crop to bbox
    path = check_path(path_str, make_dir=False)
    if not path.exists():
        raise FileNotFoundError(f"Raster file {path} does not exist.")
    if GDAL_ENV is False:
        with rasterio.open(path) as dataset:
            crs_wkt = dataset.crs.to_wkt() if dataset.crs is not None else None
            dataset_bounds = dataset.bounds
            no_data_val = dataset.nodata
            if bbox is not None:
                # Create bbox geometry for masking
                bbox_geom = geometry.box(*bbox)
                if not (
                    dataset_bounds.left <= bbox[0] <= dataset_bounds.right
                    and dataset_bounds.left <= bbox[2] <= dataset_bounds.right
                    and dataset_bounds.bottom <= bbox[1] <= dataset_bounds.top
                    and dataset_bounds.bottom <= bbox[3] <= dataset_bounds.top
                ):
                    raise ValueError("Bounding box is not fully contained within the raster dataset bounds")
                rast, trf = mask(dataset, [bbox_geom], crop=True)
            else:
                rast = dataset.read()
                trf = dataset.transform
            # Convert rasterio Affine to GDAL-style list
            trf_arr = [trf.c, trf.a, trf.b, trf.f, trf.d, trf.e]
            # rast shape: (bands, rows, cols)
            if rast.ndim == 3:
                if band < 0 or band >= rast.shape[0]:
                    raise IndexError(f"Requested band {band} out of range; raster has {rast.shape[0]} band(s)")
                rast_arr = rast[band].astype(float)
            else:
                rast_arr = rast.astype(float)
    else:
        dataset = gdal.Open(str(path))
        if dataset is None:
            raise FileNotFoundError(f"Could not open {path}")
        trf = dataset.GetGeoTransform()
        # GetProjection returns WKT string (or empty string)
        crs_wkt = dataset.GetProjection() or None
        rb = dataset.GetRasterBand(band + 1)
        if rb is None:
            dataset = None
            raise IndexError(f"Requested band {band} out of range in GDAL dataset")
        rast_arr = rb.ReadAsArray().astype(float)
        no_data_val = rb.GetNoDataValue()
        if bbox is not None:
            min_x, min_y, max_x, max_y = bbox
            xoff = int((min_x - trf[0]) / trf[1])
            yoff = int((trf[3] - max_y) / abs(trf[5]))
            xsize = int((max_x - min_x) / trf[1])
            ysize = int((max_y - min_y) / abs(trf[5]))
            # guard offsets/sizes
            if xoff < 0 or yoff < 0 or xsize <= 0 or ysize <= 0:
                dataset = None
                raise ValueError("Computed window from bbox is out of raster bounds or invalid")
            rast_arr = rast_arr[yoff : yoff + ysize, xoff : xoff + xsize]
            trf_arr = [min_x, trf[1], 0, max_y, 0, trf[5]]
        else:
            trf_arr = [trf[0], trf[1], 0, trf[3], 0, trf[5]]
        dataset = None  # ensure dataset closed
    # Handle no-data (support NaN)
    if no_data_val is not None and not np.isnan(no_data_val):
        logger.info(f"No-data value is {no_data_val}, replacing with NaN")
        rast_arr[rast_arr == no_data_val] = np.nan
    if rast_arr.size == 0:
        raise ValueError("Raster array is empty after loading/cropping")
    if rast_arr.min() < 0:
        raise ValueError("Raster contains negative values")
    return rast_arr, trf_arr, crs_wkt, no_data_val


def xy_to_lnglat(crs_wkt: str | None, x, y):
    """Convert x, y coordinates to longitude and latitude.

    Accepts scalar or array-like x/y. If crs_wkt is None the inputs are
    assumed already to be lon/lat and are returned unchanged.
    """
    if crs_wkt is None:
        logger.info("No CRS provided, assuming coordinates are already in WGS84 (lon/lat).")
        return x, y

    try:
        if GDAL_ENV is False:
            source_crs = pyproj.CRS(crs_wkt)
            target_crs = pyproj.CRS(4326)  # WGS84
            transformer = pyproj.Transformer.from_crs(source_crs, target_crs, always_xy=True)
            lng, lat = transformer.transform(x, y)
        else:
            old_cs = gdal.osr.SpatialReference()
            old_cs.ImportFromWkt(crs_wkt)
            new_cs = gdal.osr.SpatialReference()
            new_cs.ImportFromEPSG(4326)
            transform = gdal.osr.CoordinateTransformation(old_cs, new_cs)
            out = transform.TransformPoint(float(x), float(y))
            lng, lat = out[0], out[1]

        return lng, lat

    except Exception:
        logger.exception("Failed to transform coordinates")
        raise



# This is the main function of the SOLWEIG model
# 2025-Mar-21
# Fredrik Lindberg, fredrikl@gvc.gu.se
# Goteborg Urban Climate Group
# Gothenburg University

#sommon imports
from __future__ import absolute_import
from ...util.umep_solweig_export_component import read_solweig_config
from ...util.SEBESOLWEIGCommonFiles.Solweig_v2015_metdata_noload import Solweig_2015a_metdata_noload
from ...util.SEBESOLWEIGCommonFiles.clearnessindex_2013b import clearnessindex_2013b

from ...functions.SOLWEIGpython import Solweig_2025a_calc_forprocessing as so
#from ...functions.SOLWEIGpython import WriteMetadataSOLWEIG # Not needed anymore?
from ...functions.SOLWEIGpython import PET_calculations as p
from ...functions.SOLWEIGpython import UTCI_calculations as utci
from ...functions.SOLWEIGpython.CirclePlotBar import PolarBarPlot
from ...functions.SOLWEIGpython.wall_surface_temperature import load_walls
from ...functions.SOLWEIGpython.wallOfInterest import wallOfInterest
from ...functions.SOLWEIGpython.wallsAsNetCDF import walls_as_netcdf
from ...functions.SOLWEIGpython.Tgmaps_v1 import Tgmaps_v1
from ...functions import wallalgorithms as wa

import numpy as np
import json
import zipfile
import pandas as pd
import matplotlib.pylab as plt
from shutil import copyfile

# imports from osgeo/qgis dependency
try:
    from osgeo import gdal
    from osgeo.gdalconst import *
    from ...util.misc import saveraster, xy2latlon_fromraster
    from qgis.core import QgsVectorLayer, QgsRasterLayer
except:
    pass

# imports for standalone
try:
    from umep import common
    from rasterio.transform import xy, rowcol
    import pyproj
    from tqdm import tqdm
    import geopandas as gpd
except:
    pass


# import numpy as np
# from .daylen import daylen
# from ...util.SEBESOLWEIGCommonFiles.clearnessindex_2013b import clearnessindex_2013b
# from ...util.SEBESOLWEIGCommonFiles.diffusefraction import diffusefraction
# from ...util.SEBESOLWEIGCommonFiles.shadowingfunction_wallheight_13 import shadowingfunction_wallheight_13
# from ...util.SEBESOLWEIGCommonFiles.shadowingfunction_wallheight_23 import shadowingfunction_wallheight_23
# from .gvf_2018a import gvf_2018a
# from .cylindric_wedge import cylindric_wedge
# from .TsWaveDelay_2015a import TsWaveDelay_2015a
# from .Kup_veg_2015a import Kup_veg_2015a
# # from .Lside_veg_v2015a import Lside_veg_v2015a
# # from .Kside_veg_v2019a import Kside_veg_v2019a
# from .Kside_veg_v2022a import Kside_veg_v2022a
# from ...util.SEBESOLWEIGCommonFiles.Perez_v3 import Perez_v3
# from ...util.SEBESOLWEIGCommonFiles.create_patches import create_patches

def solweig_run(configPath, feedback):

    """
    Input:
    configPath : config file including geodata paths and settings.
    feedback : To communicate with qgis gui. Set to None if standalone
    """
    
    # Load config file
    configDict = read_solweig_config(configPath)

    # Load parameters settings for SOLWEIG
    with open(configDict['para_json_path'], "r") as jsn:
        param = json.load(jsn)

    standAlone = int(configDict['standalone'])

    #Load DSM
    if standAlone == 1: 
        dsm, dsm_transf, dsm_crs = common.load_raster(configDict['filepath_dsm'], bbox=None) 
        scale = 1 / dsm_transf.a
        # dsm_height, dsm_width = dsm.shape  # y rows by x cols
        # y is flipped - so return max for lower row
        minx, miny = xy(dsm_transf, dsm.shape[0], 0)
        # Define the source and target CRS
        source_crs = pyproj.CRS(dsm_crs)
        target_crs = pyproj.CRS(4326)  # WGS 84
        # Create a transformer object
        transformer = pyproj.Transformer.from_crs(source_crs, target_crs, always_xy=True)
        # Perform the transformation
        lon, lat = transformer.transform(minx, miny) 
        nd = -9999 #TODO: extract nodatavalue from rasterio
    else:
        dsmlayer = QgsRasterLayer(configDict['filepath_dsm'])
        dsm_wkt = dsmlayer.crs().toWkt()
        gdal_dsm = gdal.Open(configDict['filepath_dsm'])
        lat, lon, scale, minx, miny = xy2latlon_fromraster(dsm_wkt, gdal_dsm)
        dsm = gdal_dsm.ReadAsArray().astype(float)       
        nd = gdal_dsm.GetRasterBand(1).GetNoDataValue()

    rows = dsm.shape[0]
    cols = dsm.shape[1]
    
    # response to issue #85
    dsm[dsm == nd] = 0.
    if dsm.min() < 0:
        dsmraise = np.abs(dsm.min())
        dsm = dsm + dsmraise
    else:
        dsmraise = 0

    alt = np.median(dsm)
    if alt < 0:
        alt = 3

    #Vegetation
    transVeg = param["Tree_settings"]["Value"]["Transmissivity"] 
    trunkratio = param["Tree_settings"]["Value"]["Trunk_ratio"] 
    usevegdem = int(configDict['usevegdem'])
    if usevegdem == 1:
        if standAlone == 0:
            vegdsm = gdal.Open(configDict['filepath_cdsm']).ReadAsArray().astype(float)
        else:
            vegdsm, _ , _ = common.load_raster(configDict['filepath_cdsm'], bbox=None)
        if configDict['filepath_tdsm'] is None:
            if standAlone == 1:
                vegdsm2 = gdal.Open(configDict['filepath_tdsm']).ReadAsArray().astype(float)
            else:
                vegdsm2, _ , _ = common.load_raster(configDict['filepath_tdsm'], bbox=None)
        else:
            vegdsm2 = vegdsm * trunkratio
    else:
        vegdsm = 0
        vegdsm2 = 0

    #Land cover
    landcover = int(configDict['landcover'])
    if landcover == 1:
        if standAlone == 0:
            lcgrid = gdal.Open(configDict['filepath_lc']).ReadAsArray().astype(float)
        else:
            lcgrid, _ , _ = common.load_raster(configDict['filepath_lc'], bbox=None)
    else:
        lcgrid = 0

    #DEM for buildings #TODO: fix nodata in standalone
    demforbuild = int(configDict['demforbuild'])
    if demforbuild == 1:
        if standAlone == 0:
            gdal_dem = gdal.Open(configDict['filepath_dem'])# .ReadAsArray().astype(float)
            dem = gdal_dem.ReadAsArray().astype(float)
            nd = gdal_dem.GetRasterBand(1).GetNoDataValue()
        else:
            dem, _ , _ = common.load_raster(configDict['filepath_dem'], bbox=None)
            nd = -9999 #TODO: standAlone nd exposure

        # response to issue and #230
        dem[dem == nd] = 0.
        if dem.min() < 0:
            demraise = np.abs(dem.min())
            dem = dem + demraise
        else:
            demraise = 0

    #SVF
    zip = zipfile.ZipFile(configDict['input_svf'], 'r')
    zip.extractall(configDict['working_dir'])
    zip.close()

    if standAlone == 0:
        svf = gdal.Open(configDict['working_dir'] + "/svf.tif").ReadAsArray().astype(float)
        svfN = gdal.Open(configDict['working_dir'] + "/svfN.tif").ReadAsArray().astype(float)
        svfS = gdal.Open(configDict['working_dir'] + "/svfS.tif").ReadAsArray().astype(float)
        svfE = gdal.Open(configDict['working_dir'] + "/svfE.tif").ReadAsArray().astype(float)
        svfW = gdal.Open(configDict['working_dir'] + "/svfW.tif").ReadAsArray().astype(float)
    else:
        svf, _ , _ = common.load_raster(configDict['working_dir'] + "/svf.tif", bbox=None)
        svfN, _ , _ = common.load_raster(configDict['working_dir'] + "/svfN.tif", bbox=None)
        svfS, _ , _ = common.load_raster(configDict['working_dir'] + "/svfS.tif", bbox=None)
        svfE, _ , _ = common.load_raster(configDict['working_dir'] + "/svfE.tif", bbox=None)
        svfW, _ , _ = common.load_raster(configDict['working_dir'] + "/svfW.tif", bbox=None)    

    if usevegdem == 1:
        if standAlone == 0:
            svfveg = gdal.Open(configDict['working_dir'] + "/svfveg.tif").ReadAsArray().astype(float)
            svfNveg = gdal.Open(configDict['working_dir'] + "/svfNveg.tif").ReadAsArray().astype(float)
            svfSveg = gdal.Open(configDict['working_dir'] + "/svfSveg.tif").ReadAsArray().astype(float)
            svfEveg = gdal.Open(configDict['working_dir'] + "/svfEveg.tif").ReadAsArray().astype(float)
            svfWveg = gdal.Open(configDict['working_dir'] + "/svfWveg.tif").ReadAsArray().astype(float)

            svfaveg = gdal.Open(configDict['working_dir'] + "/svfaveg.tif").ReadAsArray().astype(float)
            svfNaveg = gdal.Open(configDict['working_dir'] + "/svfNaveg.tif").ReadAsArray().astype(float)
            svfSaveg = gdal.Open(configDict['working_dir'] + "/svfSaveg.tif").ReadAsArray().astype(float)
            svfEaveg = gdal.Open(configDict['working_dir'] + "/svfEaveg.tif").ReadAsArray().astype(float)
            svfWaveg = gdal.Open(configDict['working_dir'] + "/svfWaveg.tif").ReadAsArray().astype(float)
        else:
            svfveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfveg.tif", bbox=None)
            svfNveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfNveg.tif", bbox=None)
            svfSveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfSveg.tif", bbox=None)
            svfEveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfEveg.tif", bbox=None)
            svfWveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfWveg.tif", bbox=None)

            svfaveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfaveg.tif", bbox=None)
            svfNaveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfNaveg.tif", bbox=None)
            svfSaveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfSaveg.tif", bbox=None)
            svfEaveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfEaveg.tif", bbox=None)
            svfWaveg, _ , _ = common.load_raster(configDict['working_dir'] + "/svfWaveg.tif", bbox=None)
    else:
        svfveg = np.ones((rows, cols))
        svfNveg = np.ones((rows, cols))
        svfSveg = np.ones((rows, cols))
        svfEveg = np.ones((rows, cols))
        svfWveg = np.ones((rows, cols))
        svfaveg = np.ones((rows, cols))
        svfNaveg = np.ones((rows, cols))
        svfSaveg = np.ones((rows, cols))
        svfEaveg = np.ones((rows, cols))
        svfWaveg = np.ones((rows, cols))

    tmp = svf + svfveg - 1.
    tmp[tmp < 0.] = 0.
    # %matlab crazyness around 0
    svfalfa = np.arcsin(np.exp((np.log((1. - tmp)) / 2.)))

    if standAlone == 0:
        wallheight = gdal.Open(configDict['filepath_wh']).ReadAsArray().astype(float)
        wallaspect = gdal.Open(configDict['filepath_wa']).ReadAsArray().astype(float)
    else:
        wallheight, _ , _ = common.load_raster(configDict['filepath_wh'], bbox=None)
        wallaspect, _ , _ = common.load_raster(configDict['filepath_wa'], bbox=None)

    # Metdata
    headernum = 1
    delim = ' '
    Twater = []

    metdata = np.loadtxt(configDict['input_met'],skiprows=headernum, delimiter=delim) 

    location = {'longitude': lon, 'latitude': lat, 'altitude': alt}
    YYYY, altitude, azimuth, zen, jday, leafon, dectime, altmax = Solweig_2015a_metdata_noload(metdata,location, int(configDict['utc']))

    DOY = metdata[:, 1]
    hours = metdata[:, 2]
    minu = metdata[:, 3]
    Ta = metdata[:, 11]
    RH = metdata[:, 10]
    radG = metdata[:, 14]
    radD = metdata[:, 21]
    radI = metdata[:, 22]
    P = metdata[:, 12]
    Ws = metdata[:, 9]

    # POIs check
    if configDict['poi_file'] is not '': # usePOI:
        header = 'yyyy id   it imin dectime altitude azimuth kdir kdiff kglobal kdown   kup    keast ksouth ' \
                    'kwest knorth ldown   lup    least lsouth lwest  lnorth   Ta      Tg     RH    Esky   Tmrt    ' \
                    'I0     CI   Shadow  SVF_b  SVF_bv KsideI PET UTCI  CI_Tg   CI_TgG  KsideD  Lside   diffDown    Kside'
        poi_field = configDict['poi_field'] #self.parameterAsString(parameters, self.POI_FIELD, context)
        poiname = []
        if standAlone == 0:
            vlayer = QgsVectorLayer(configDict['poi_file'], 'point', 'ogr')
            idx = vlayer.fields().indexFromName(poi_field)
            numfeat = vlayer.featureCount()
            poisxy = np.zeros((numfeat, 3)) - 999
            ind = 0
            for f in vlayer.getFeatures():  # looping through each POI
                y = f.geometry().centroid().asPoint().y()
                x = f.geometry().centroid().asPoint().x()
                poiname.append(f.attributes()[idx])
                poisxy[ind, 0] = ind
                poisxy[ind, 1] = np.round((x - minx) * scale)
                if miny >= 0:
                    poisxy[ind, 2] = np.round((miny + rows * (1. / scale) - y) * scale)
                else:
                    poisxy[ind, 2] = np.round((miny + rows * (1. / scale) - y) * scale)
                ind += 1
        else:
            pois_gdf = gpd.read_file(configDict['poi_file'])
            numfeat = pois_gdf.shape[0]
            poisxy = np.zeros((numfeat, 3)) - 999
            for idx, row in pois_gdf.iterrows():
                y, x = rowcol(dsm_transf, row["geometry"].centroid.x, row["geometry"].centroid.y) #TODO: This produce different result since no standalone round coordinates
                poiname.append(row[configDict['poi_field']])
                poisxy[idx, 0] = idx
                poisxy[idx, 1] = x
                poisxy[idx, 2] = y

        for k in range(0, poisxy.shape[0]):
            poi_save = []  # np.zeros((1, 33))
            data_out = configDict['output_dir'] + '/POI_' + str(poiname[k]) + '.txt'
            np.savetxt(data_out, poi_save,  delimiter=' ', header=header, comments='')
        print(poisxy)
        # Num format for POI output
        numformat = '%d %d %d %d %.5f ' + '%.2f ' * 36

        # Other PET variables
        sensorheight = param['Wind_Height']['Value']['magl']
        age = param['PET_settings']['Value']['Age']
        mbody = param['PET_settings']['Value']['Weight']
        ht = param['PET_settings']['Value']['Height']
        clo = param['PET_settings']['Value']['clo']
        activity = param['PET_settings']['Value']['Activity']
        sex = param['PET_settings']['Value']['Sex']
    else:
        poisxy = None
    
    #Posture settings
    if  param['Tmrt_params']['Value']['posture'] == "Standing":
        Fside = param["Posture"]["Standing"]["Value"]["Fside"]
        Fup = param["Posture"]["Standing"]["Value"]["Fup"]
        height = param["Posture"]["Standing"]["Value"]["height"]
        Fcyl = param["Posture"]["Standing"]["Value"]["Fcyl"] 
        pos = 1
    else:
        Fside = param["Posture"]["Sitting"]["Value"]["Fside"]
        Fup = param["Posture"]["Sitting"]["Value"]["Fup"]
        height = param["Posture"]["Sitting"]["Value"]["height"]
        Fcyl = param["Posture"]["Sitting"]["Value"]["Fcyl"]
        pos = 0

    # Radiative surface influence, Rule of thumb by Schmid et al. (1990).
    first = np.round(height)
    if first == 0.:
        first = 1.
    second = np.round((height * 20.))

    if usevegdem == 1:
        # Conifer or deciduous
        if configDict['conifer_bool']:
            leafon = np.ones((1, DOY.shape[0]))
        else:
            leafon = np.zeros((1, DOY.shape[0]))
            if param["Tree_settings"]["Value"]["First_day_leaf"] > param["Tree_settings"]["Value"]["Last_day_leaf"]:
                leaf_bool = ((DOY > param["Tree_settings"]["Value"]["First_day_leaf"]) | (DOY < param["Tree_settings"]["Value"]["Last_day_leaf"]))
            else:
                leaf_bool = ((DOY > param["Tree_settings"]["Value"]["First_day_leaf"]) & (DOY < param["Tree_settings"]["Value"]["Last_day_leaf"]))
            leafon[0, leaf_bool] = 1

        # % Vegetation transmittivity of shortwave radiation
        psi = leafon * transVeg
        psi[leafon == 0] = 0.5
        # amaxvalue
        vegmax = vegdsm.max()
        amaxvalue = dsm.max() - dsm.min()
        amaxvalue = np.maximum(amaxvalue, vegmax)

        # Elevation vegdsms if buildingDEM includes ground heights
        vegdsm = vegdsm + dsm
        vegdsm[vegdsm == dsm] = 0
        vegdsm2 = vegdsm2 + dsm
        vegdsm2[vegdsm2 == dsm] = 0

        # % Bush separation
        bush = np.logical_not((vegdsm2 * vegdsm)) * vegdsm

        svfbuveg = (svf - (1. - svfveg) * (1. - transVeg))  # % major bug fixed 20141203
    else:
        psi = leafon * 0. + 1.
        svfbuveg = svf
        bush = np.zeros([rows, cols])
        amaxvalue = 0

    # Initialization of maps
    Knight = np.zeros((rows, cols))
    Tgmap1 = np.zeros((rows, cols))
    Tgmap1E = np.zeros((rows, cols))
    Tgmap1S = np.zeros((rows, cols))
    Tgmap1W = np.zeros((rows, cols))
    Tgmap1N = np.zeros((rows, cols))

    # Create building boolean raster from either land cover or height rasters
    if demforbuild == 0:
        buildings = np.copy(lcgrid)
        buildings[buildings == 7] = 1
        buildings[buildings == 6] = 1
        buildings[buildings == 5] = 1
        buildings[buildings == 4] = 1
        buildings[buildings == 3] = 1
        buildings[buildings == 2] = 0
    else:
        buildings = dsm - dem
        buildings[buildings < 2.] = 1.
        buildings[buildings >= 2.] = 0.

    if int(configDict['savebuild']) == 1:
        if standAlone == 0:
            saveraster(gdal_dsm, configDict['output_dir'] + '/buildings.tif', buildings)
        else:
            common.save_raster(configDict['output_dir'] + '/buildings.tif', buildings, dsm_transf, dsm_crs)

    # Import shadow matrices (Anisotropic sky)
    anisotropic_sky = int(configDict['aniso'])
    if anisotropic_sky == 1:  #UseAniso
        data = np.load(configDict['input_aniso'])
        shmat = data['shadowmat']
        vegshmat = data['vegshadowmat']
        vbshvegshmat = data['vbshmat']
        if usevegdem == 1:
            diffsh = np.zeros((rows, cols, shmat.shape[2]))
            for i in range(0, shmat.shape[2]):
                diffsh[:, :, i] = shmat[:, :, i] - (1 - vegshmat[:, :, i]) * (1 - transVeg) # changes in psi not implemented yet
        else:
            diffsh = shmat

        # Estimate number of patches based on shadow matrices
        if shmat.shape[2] == 145:
            patch_option = 1 # patch_option = 1 # 145 patches
        elif shmat.shape[2] == 153:
            patch_option = 2 # patch_option = 2 # 153 patches
        elif shmat.shape[2] == 306:
            patch_option = 3 # patch_option = 3 # 306 patches
        elif shmat.shape[2] == 612:
            patch_option = 4 # patch_option = 4 # 612 patches

        # asvf to calculate sunlit and shaded patches
        asvf = np.arccos(np.sqrt(svf))

        # Empty array for steradians
        steradians = np.zeros((shmat.shape[2]))
    else:
        # anisotropic_sky = 0
        diffsh = None
        shmat = None
        vegshmat = None
        vbshvegshmat = None
        asvf = None
        patch_option = 0
        steradians = 0
    
    # % Ts parameterisation maps
    if landcover == 1.:
        # Get land cover properties for Tg wave (land cover scheme based on Bogren et al. 2000, explained in Lindberg et al., 2008 and Lindberg, Onomura & Grimmond, 2016)
        [TgK, Tstart, alb_grid, emis_grid, TgK_wall, Tstart_wall, TmaxLST, TmaxLST_wall] = Tgmaps_v1(lcgrid.copy(), param)
    else:
        TgK = Knight + param['Ts_deg']['Value']['Cobble_stone_2014a']
        Tstart = Knight - param['Tstart']['Value']['Cobble_stone_2014a']
        TmaxLST = param['TmaxLST']['Value']['Cobble_stone_2014a']
        alb_grid = Knight + param['Albedo']['Effective']['Value']['Cobble_stone_2014a']
        emis_grid = Knight + param['Emissivity']['Value']['Cobble_stone_2014a']
        TgK_wall = param['Ts_deg']['Value']['Walls']
        Tstart_wall = param['Tstart']['Value']['Walls']
        TmaxLST_wall = param['TmaxLST']['Value']['Walls']

    # Import data for wall temperature parameterization TODO: fix for standalone
    wallScheme = int(configDict['wallscheme'])
    if wallScheme == 1:
        wallData = np.load(configDict['input_wall'])
        voxelMaps = wallData['voxelId']
        voxelTable = wallData['voxelTable']
        # Get wall type from standalone
        if standAlone == 1:
            wall_type_standalone = {'Brick_wall': '100', 'Concrete_wall': '101', 'Wood_wall': '102'}
            wall_type = wall_type_standalone[configDict['walltype']]
        else:
            # Get wall type set in GUI
            wall_type = configDict['walltype'] # str(100 + int(self.parameterAsString(parameters, self.WALL_TYPE, context))) #TODO

        # Get heights of walls including corners
        walls_scheme = wa.findwalls_sp(dsm, 2, np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]]))
        # Get aspects of walls including corners
        dirwalls_scheme = wa.filter1Goodwin_as_aspect_v3(walls_scheme.copy(), scale, dsm, None, 100./180.)
        
        # Used in wall temperature parameterization scheme
        first_timestep = (pd.to_datetime(YYYY[0][0], format='%Y') + pd.to_timedelta(DOY[0] - 1, unit='d') + 
            pd.to_timedelta(hours[0], unit='h') + pd.to_timedelta(minu[0], unit='m'))
        second_timestep = (pd.to_datetime(YYYY[0][1], format='%Y') + pd.to_timedelta(DOY[1] - 1, unit='d') + 
            pd.to_timedelta(hours[1], unit='h') + pd.to_timedelta(minu[1], unit='m'))
        
        timeStep = (second_timestep - first_timestep).seconds

        # Load voxelTable as Pandas DataFrame
        voxelTable, dirwalls_scheme = load_walls(voxelTable, param, wall_type, dirwalls_scheme, Ta[0], timeStep, alb_grid, landcover, lcgrid, dsm)
        # Unique wall types
        #thermal_effusivity = voxelTable['thermalEffusivity'].unique()

        # Empty wall temperature matrix for parameterization scheme
        # wallScheme_feedback = "Running with wall parameterization scheme. Walls divided into " + str(int(voxelTable.shape[0])) + " unique voxels."
        # feedback.setProgressText(wallScheme_feedback)

        # Use wall of interest
        woi_file = configDict['woi_file']
        if woi_file:
            (dsm_minx, dsm_x_size, dsm_x_rotation, dsm_miny, dsm_y_rotation, dsm_y_size) = gdal_dsm.GetGeoTransform() #TODO: fix for standalone

            woi_field = configDict['woi_field'] #self.parameterAsStrings(parameters, self.WOI_FIELD, context)
            woisxy, woiname = wallOfInterest(configDict['woi_file'], woi_field, minx, miny, scale, rows, configDict['output_dir'], dsm_minx, dsm_x_size, dsm_miny, dsm_y_size)

        # Create pandas datetime object to be used when createing an xarray DataSet where wall temperatures/radiation is stored and eventually saved as a NetCDf
        if configDict["wallnetcdf"] == 1:
            met_for_xarray = (pd.to_datetime(YYYY[0][:], format='%Y') + pd.to_timedelta(DOY - 1, unit='d') + 
                            pd.to_timedelta(hours, unit='h') + pd.to_timedelta(minu, unit='m'))
    else:
        wallScheme = 0
        voxelMaps = 0
        voxelTable = 0
        timeStep = 0
        #thermal_effusivity = 0
        walls_scheme = np.ones((rows, cols)) * 10.
        dirwalls_scheme = np.ones((rows, cols)) * 10.

    # Initialisation of time related variables
    if Ta.__len__() == 1:
        timestepdec = 0
    else:
        timestepdec = dectime[1] - dectime[0]
    timeadd = 0.
    # timeaddE = 0.
    # timeaddS = 0.
    # timeaddW = 0.
    # timeaddN = 0.
    firstdaytime = 1.


    # Save hemispheric image
    if anisotropic_sky == 1: 
        if not poisxy is None:
            patch_characteristics = np.zeros((shmat.shape[2], poisxy.shape[0]))
            for idx in range(poisxy.shape[0]):
                for idy in range(shmat.shape[2]):
                    # Calculations for patches on sky, shmat = 1 = sky is visible
                    temp_sky = ((shmat[:,:,idy] == 1) & (vegshmat[:,:,idy] == 1))
                    # Calculations for patches that are vegetation, vegshmat = 0 = shade from vegetation
                    temp_vegsh = ((vegshmat[:,:,idy] == 0) | (vbshvegshmat[:,:,idy] == 0))
                    # Calculations for patches that are buildings, shmat = 0 = shade from buildings
                    temp_vbsh = (1 - shmat[:,:,idy]) * vbshvegshmat[:,:,idy]
                    temp_sh = (temp_vbsh == 1)
                    if wallScheme:
                        temp_sh_w = temp_sh * voxelMaps[:, :, idy]
                        temp_sh_roof = temp_sh * (voxelMaps[:, :, idy] == 0)
                    else:
                        temp_sh_w = 0
                        temp_sh_roof = 0
                    # Sky patch
                    if temp_sky[int(poisxy[idx, 2]), int(poisxy[idx, 1])]:
                        patch_characteristics[idy,idx] = 1.8
                    # Vegetation patch
                    elif (temp_vegsh[int(poisxy[idx, 2]), int(poisxy[idx, 1])]):
                        patch_characteristics[idy,idx] = 2.5
                    # Building patch
                    elif (temp_sh[int(poisxy[idx, 2]), int(poisxy[idx, 1])]):
                        if wallScheme:
                            if (temp_sh_w[int(poisxy[idx, 2]), int(poisxy[idx, 1])]):
                                patch_characteristics[idy, idx] = 4.5
                            elif (temp_sh_roof[int(poisxy[idx, 2]), int(poisxy[idx, 1])]):
                                patch_characteristics[idy, idx] = 6.0
                        else:
                            patch_characteristics[idy,idx] = 4.5
                    # Roof patch
                    #elif

    # If metfile starts at night
    CI = 1.

    # reading variables from config and parameters that is not yet presented
    cyl = int(configDict["cyl"])
    albedo_b = param['Albedo']['Effective']['Value']['Walls']
    ewall = param['Emissivity']['Value']['Walls']
    onlyglobal = int(configDict['onlyglobal'])
    elvis = 0.0
    absK = param['Tmrt_params']['Value']['absK'] 
    absL = param['Tmrt_params']['Value']['absL'] 
    
    # Main loop
    tmrtplot = np.zeros((rows, cols))
    TgOut1 = np.zeros((rows, cols))

    # Initiate array for I0 values
    if np.unique(DOY).shape[0] > 1:
        unique_days = np.unique(DOY)
        first_unique_day = DOY[DOY == unique_days[0]]
        I0_array = np.zeros((first_unique_day.shape[0]))
    else:
        first_unique_day = DOY.copy()
        I0_array = np.zeros((DOY.shape[0]))

    if standAlone == 1:
        progress = tqdm(total=Ta.__len__())

    for i in np.arange(0, Ta.__len__()):
        if feedback is not None:
            feedback.setProgress(int(i * (100. / Ta.__len__()))) # move progressbar forward
            if feedback.isCanceled():
                feedback.setProgressText("Calculation cancelled")
                break
        else:
            progress.update(1)

        # Daily water body temperature
        if landcover == 1:
            if ((dectime[i] - np.floor(dectime[i]))) == 0 or (i == 0):
                Twater = np.mean(Ta[jday[0] == np.floor(dectime[i])])
        # Nocturnal cloudfraction from Offerle et al. 2003
        if (dectime[i] - np.floor(dectime[i])) == 0:
            daylines = np.where(np.floor(dectime) == dectime[i])
            if daylines.__len__() > 1:
                alt = altitude[0][daylines]
                alt2 = np.where(alt > 1)
                rise = alt2[0][0]
                [_, CI, _, _, _] = clearnessindex_2013b(zen[0, i + rise + 1], jday[0, i + rise + 1],
                                                        Ta[i + rise + 1],
                                                        RH[i + rise + 1] / 100., radG[i + rise + 1], location,
                                                        P[i + rise + 1])
                if (CI > 1.) or (CI == np.inf):
                    CI = 1.
            else:
                CI = 1.

        # Only if Kdir is derived from horizontal global shortwave and horizontal diffuse shortwave
        # if altitude[0][i] > 0:
        #     radI[i] = radI[i]/np.sin(altitude[0][i] * np.pi/180)
        # else:
        #     radG[i] = 0.
        #     radD[i] = 0.
        #     radI[i] = 0.


        Tmrt, Kdown, Kup, Ldown, Lup, Tg, ea, esky, I0, CI, shadow, firstdaytime, timestepdec, timeadd, \
                Tgmap1, Tgmap1E, Tgmap1S, Tgmap1W, Tgmap1N, Keast, Ksouth, Kwest, Knorth, Least, \
                Lsouth, Lwest, Lnorth, KsideI, TgOut1, TgOut, radIout, radDout, \
                Lside, Lsky_patch_characteristics, CI_Tg, CI_TgG, KsideD, \
                    dRad, Kside, steradians, voxelTable = so.Solweig_2025a_calc(
                    i, dsm, scale, rows, cols, svf, svfN, svfW, svfE, svfS, svfveg,
                    svfNveg, svfEveg, svfSveg, svfWveg, svfaveg, svfEaveg, svfSaveg, svfWaveg, svfNaveg, \
                    vegdsm, vegdsm2, albedo_b, absK, absL, ewall, Fside, Fup, Fcyl, altitude[0][i],
                    azimuth[0][i], zen[0][i], jday[0][i], usevegdem, onlyglobal, buildings, location,
                    psi[0][i], landcover, lcgrid, dectime[i], altmax[0][i], wallaspect,
                    wallheight, cyl, elvis, Ta[i], RH[i], radG[i], radD[i], radI[i], P[i], amaxvalue,
                    bush, Twater, TgK, Tstart, alb_grid, emis_grid, TgK_wall, Tstart_wall, TmaxLST,
                    TmaxLST_wall, first, second, svfalfa, svfbuveg, firstdaytime, timeadd, timestepdec, 
                    Tgmap1, Tgmap1E, Tgmap1S, Tgmap1W, Tgmap1N, CI, TgOut1, diffsh, shmat, vegshmat, vbshvegshmat, 
                    anisotropic_sky, asvf, patch_option, voxelMaps, voxelTable, Ws[i], wallScheme, timeStep, steradians, walls_scheme, dirwalls_scheme)

        # Save I0 for I0 vs. Kdown output plot to check if UTC is off
        if i < first_unique_day.shape[0]:
            I0_array[i] = I0
        elif i == first_unique_day.shape[0]:
            # Output I0 vs. Kglobal plot
            radG_for_plot = radG[DOY == first_unique_day[0]]
            # hours_for_plot = hours[DOY == first_unique_day[0]]
            dectime_for_plot = dectime[DOY == first_unique_day[0]]
            fig, ax = plt.subplots()
            ax.plot(dectime_for_plot, I0_array, label='I0')
            ax.plot(dectime_for_plot, radG_for_plot, label='Kglobal')
            ax.set_ylabel('Shortwave radiation [$Wm^{-2}$]')
            ax.set_xlabel('Decimal time')
            ax.set_title('UTC' + str(configDict['utc']))
            ax.legend()
            fig.savefig(configDict['output_dir'] + '/metCheck.png', dpi=150)                

        tmrtplot = tmrtplot + Tmrt

        if altitude[0][i] > 0:
            w = 'D'
        else:
            w = 'N'

        # Write to POIs
        if not poisxy is None:              
            for k in range(0, poisxy.shape[0]):
                poi_save = np.zeros((1, 41))
                poi_save[0, 0] = YYYY[0][i]
                poi_save[0, 1] = jday[0][i]
                poi_save[0, 2] = hours[i]
                poi_save[0, 3] = minu[i]
                poi_save[0, 4] = dectime[i]
                poi_save[0, 5] = altitude[0][i]
                poi_save[0, 6] = azimuth[0][i]
                poi_save[0, 7] = radIout
                poi_save[0, 8] = radDout
                poi_save[0, 9] = radG[i]
                poi_save[0, 10] = Kdown[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 11] = Kup[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 12] = Keast[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 13] = Ksouth[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 14] = Kwest[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 15] = Knorth[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 16] = Ldown[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 17] = Lup[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 18] = Least[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 19] = Lsouth[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 20] = Lwest[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 21] = Lnorth[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 22] = Ta[i]
                poi_save[0, 23] = TgOut[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 24] = RH[i]
                poi_save[0, 25] = esky
                poi_save[0, 26] = Tmrt[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 27] = I0
                poi_save[0, 28] = CI
                poi_save[0, 29] = shadow[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 30] = svf[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 31] = svfbuveg[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 32] = KsideI[int(poisxy[k, 2]), int(poisxy[k, 1])]
                # Recalculating wind speed based on powerlaw
                WsPET = (1.1 / sensorheight) ** 0.2 * Ws[i]
                WsUTCI = (10. / sensorheight) ** 0.2 * Ws[i]
                resultPET = p._PET(Ta[i], RH[i], Tmrt[int(poisxy[k, 2]), int(poisxy[k, 1])], WsPET,
                                    mbody, age, ht, activity, clo, sex)
                poi_save[0, 33] = resultPET
                resultUTCI = utci.utci_calculator(Ta[i], RH[i], Tmrt[int(poisxy[k, 2]), int(poisxy[k, 1])],
                                                    WsUTCI)
                poi_save[0, 34] = resultUTCI
                poi_save[0, 35] = CI_Tg
                poi_save[0, 36] = CI_TgG
                poi_save[0, 37] = KsideD[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 38] = Lside[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 39] = dRad[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 40] = Kside[int(poisxy[k, 2]), int(poisxy[k, 1])]
                data_out = configDict['output_dir'] + '/POI_' + str(poiname[k]) + '.txt'
                # f_handle = file(data_out, 'a')
                f_handle = open(data_out, 'ab')
                np.savetxt(f_handle, poi_save, fmt=numformat)
                f_handle.close()

        # If wall temperature parameterization scheme is in use
        if configDict['wallscheme'] == 1: # folderWallScheme: TODO: Fix for standalone
            # Store wall data for output
            if not woisxy is None:
                for k in range(0, woisxy.shape[0]):
                    temp_wall = voxelTable.loc[((voxelTable['ypos'] == woisxy[k, 2]) & (voxelTable['xpos'] == woisxy[k, 1])), 'wallTemperature'].to_numpy()
                    K_in = voxelTable.loc[((voxelTable['ypos'] == woisxy[k, 2]) & (voxelTable['xpos'] == woisxy[k, 1])), 'K_in'].to_numpy()
                    L_in = voxelTable.loc[((voxelTable['ypos'] == woisxy[k, 2]) & (voxelTable['xpos'] == woisxy[k, 1])), 'L_in'].to_numpy()
                    wallShade = voxelTable.loc[((voxelTable['ypos'] == woisxy[k, 2]) & (voxelTable['xpos'] == woisxy[k, 1])), 'wallShade'].to_numpy()
                    temp_all = np.concatenate([temp_wall, K_in, L_in, wallShade])
                    # temp_all = np.concatenate([temp_wall])
                    # wall_data = np.zeros((1, 7 + temp_wall.shape[0]))
                    wall_data = np.zeros((1, 7 + temp_all.shape[0]))
                    # Part of file name (wallid), i.e. WOI_wallid.txt
                    data_out = configDict['output_dir'] + '/WOI_' + str(woiname[k]) + '.txt'                    
                    if i == 0:
                        # Output file header
                        #header = 'yyyy id   it imin dectime Ta  SVF Ts'
                        header = 'yyyy id   it imin dectime Ta  SVF' + ' Ts' * temp_wall.shape[0] + ' Kin' * K_in.shape[0] + ' Lin' * L_in.shape[0] + ' shade' * wallShade.shape[0]
                        # Part of file name (wallid), i.e. WOI_wallid.txt
                        # woiname = voxelTable.loc[((voxelTable['ypos'] == woisxy[k, 2]) & (voxelTable['xpos'] == woisxy[k, 1])), 'wallId'].to_numpy()[0]
                        woi_save = []  # 
                        np.savetxt(data_out, woi_save,  delimiter=' ', header=header, comments='')                        
                    # Fill wall_data with variables
                    wall_data[0, 0] = YYYY[0][i] 
                    wall_data[0, 1] = jday[0][i]
                    wall_data[0, 2] = hours[i]
                    wall_data[0, 3] = minu[i]
                    wall_data[0, 4] = dectime[i]
                    wall_data[0, 5] = Ta[i]
                    wall_data[0, 6] = svf[int(woisxy[k, 2]), int(woisxy[k, 1])]
                    wall_data[0, 7:] = temp_all

                    # Num format for output file data
                    woi_numformat = '%d %d %d %d %.5f %.2f %.2f' + ' %.2f' * temp_all.shape[0]
                    # Open file, add data, save
                    f_handle = open(data_out, 'ab')
                    np.savetxt(f_handle, wall_data, fmt=woi_numformat)
                    f_handle.close()                

            # Save wall temperature/radiation as NetCDF TODO: fix for standAlone?
            if configDict['wallnetcdf'] == '1': # wallNetCDF:
                netcdf_output = configDict['outputDir'] + '/walls.nc'
                walls_as_netcdf(voxelTable, rows, cols, met_for_xarray, i, dsm, configDict['filepath_dsm'], netcdf_output)

        if hours[i] < 10:
            XH = '0'
        else:
            XH = ''
        if minu[i] < 10:
            XM = '0'
        else:
            XM = ''

        time_code = (str(int(YYYY[0, i])) + "_" + str(int(DOY[i])) + "_" + XH + str(int(hours[i])) + XM + str(int(minu[i])) + w)

        if configDict['outputtmrt'] == '1':
            if standAlone == 0:
                saveraster(gdal_dsm, configDict['output_dir'] + '/Tmrt_' + time_code + '.tif', Tmrt)
            else:
                common.save_raster(configDict['output_dir'] + '/Tmrt_' + time_code + '.tif', Tmrt, dsm_transf, dsm_crs)
        if configDict['outputkup'] == '1':
            if standAlone == 0:           
                saveraster(gdal_dsm, configDict['output_dir'] + '/Kup_' + time_code + '.tif', Kup)
            else:
                common.save_raster(configDict['output_dir'] + '/Kup_' + time_code + '.tif', Kup, dsm_transf, dsm_crs)
        if configDict['outputkdown'] == '1':
            if standAlone == 0:
                saveraster(gdal_dsm, configDict['output_dir'] + '/Kdown_' + time_code + '.tif', Kdown)
            else:
                common.save_raster(configDict['output_dir'] + '/Kdown_' + time_code + '.tif', Kdown, dsm_transf, dsm_crs)
        if configDict['outputlup'] == '1':
            if standAlone == 0:
                saveraster(gdal_dsm, configDict['output_dir'] + '/Lup_' + time_code + '.tif', Lup)
            else:
                common.save_raster(configDict['output_dir'] + '/Lup_' + time_code + '.tif', Lup, dsm_transf, dsm_crs)
        if configDict['outputldown'] == '1':
            if standAlone == 0:
                saveraster(gdal_dsm, configDict['output_dir'] + '/Ldown_' + time_code + '.tif', Ldown)
            else:
                common.save_raster(configDict['output_dir'] + '/Ldown_' + time_code + '.tif', Ldown, dsm_transf, dsm_crs)
        if configDict['outputsh'] == '1':
            if standAlone == 0:
                saveraster(gdal_dsm, configDict['output_dir'] + '/Shadow_' + time_code + '.tif', shadow)
            else:
                common.save_raster(configDict['output_dir'] + '/Shadow_' + time_code + '.tif', shadow, dsm_transf, dsm_crs)
            
        if configDict['outputkdiff'] == '1':
            if standAlone == 0:
                saveraster(gdal_dsm, configDict['output_dir'] + '/Kdiff_' + time_code + '.tif', dRad)  
            else:
                common.save_raster(configDict['output_dir'] + '/Kdiff_' + time_code + '.tif', dRad, dsm_transf, dsm_crs)         

        # Sky view image of patches
        if ((anisotropic_sky == 1) & (i == 0) & (not poisxy is None)):
                for k in range(poisxy.shape[0]):
                    Lsky_patch_characteristics[:,2] = patch_characteristics[:,k]
                    skyviewimage_out = configDict['output_dir'] + '/POI_' + str(poiname[k]) + '.png'
                    PolarBarPlot(Lsky_patch_characteristics, altitude[0][i], azimuth[0][i], 'Hemisphere partitioning', skyviewimage_out, 0, 5, 0)

    # Save files for Tree Planter
    if configDict['outputtreeplanter'] == '1': # outputTreeplanter:
        if feedback is not None:
            feedback.setProgressText("Saving files for Tree Planter tool")
        # Save DSM
        copyfile(configDict['filepath_dsm'], configDict['output_dir'] + '/DSM.tif')

        # Save CDSM
        if usevegdem == 1:
            copyfile(configDict['filepath_cdsm'],  configDict['output_dir'] + '/CDSM.tif')

        albedo_g = param['Albedo']['Effective']['Value']['Cobble_stone_2014a']
        eground = param['Emissivity']['Value']['Cobble_stone_2014a']

        # Saving settings from SOLWEIG for SOLWEIG1D in TreePlanter
        settingsHeader = 'UTC, posture, onlyglobal, landcover, anisotropic, cylinder, albedo_walls, albedo_ground, emissivity_walls, emissivity_ground, absK, absL, elevation, patch_option'
        settingsFmt = '%i', '%i', '%i', '%i', '%i', '%i', '%1.2f', '%1.2f', '%1.2f', '%1.2f', '%1.2f', '%1.2f', '%1.2f', '%i'
        settingsData = np.array([[int(configDict['utc']), pos, onlyglobal, landcover, anisotropic_sky, cyl, albedo_b, albedo_g, ewall, eground, absK, absL, alt, patch_option]])
        # print(settingsData)
        np.savetxt(configDict['output_dir'] + '/treeplantersettings.txt', settingsData, fmt=settingsFmt, header=settingsHeader, delimiter=' ')

    # Copying met file for SpatialTC
    copyfile(configDict['input_met'], configDict['output_dir'] + '/metforcing.txt')
    
    tmrtplot = tmrtplot / Ta.__len__()  # fix average Tmrt instead of sum, 20191022
    if standAlone == 0:
        saveraster(gdal_dsm, configDict['output_dir'] + '/Tmrt_average.tif', tmrtplot)
    else:
        common.save_raster(configDict['output_dir'] + '/Tmrt_average.tif', tmrtplot, dsm_transf, dsm_crs)
